"""Contains methods which deals with logging."""


import logging
import logging.config
from typing import Any, Dict, Optional

import yaml
from bout_runners.utils.paths import get_log_file_path, get_logger_config_path


# NOTE: Looks like mypy has trouble with recursive objects, thus this using Any looks
#       like a good solution for now
#       See also
#       https://github.com/python/typing/issues/182
def get_log_config() -> Dict[str, Any]:
    """
    Get the logging configuration.

    Returns
    -------
    config : dict
        A dictionary containing the logging configuration
    """
    log_config_path = get_logger_config_path()

    with log_config_path.open("r") as config_file:
        config: Dict[str, Any] = yaml.safe_load(config_file.read())
    return config


def set_up_logger(config: Optional[Dict[str, Any]] = None) -> None:
    """
    Set up the logger.

    Parameters
    ----------
    config : None or dict
        A dictionary containing the logging configuration
    """
    if config is None:
        config = get_log_config()
    config["handlers"]["file_handler"]["filename"] = str(
        get_log_file_path(name="bout_runners.log")
    )
    logging.config.dictConfig(config)



"""Contains the submitter factory."""


import logging

from bout_runners.submitter.local_submitter import LocalSubmitter


def get_submitter(name: str, *args, **kwargs) -> LocalSubmitter:
    """
    Return a Submitter object.

    Parameters
    ----------
    name : str
        Name of the submitter to use
    args : tuple
        Positional arguments (see the different implementation for details)
    kwargs : dict
        Keyword arguments (see the different implementation for details)

    Returns
    -------
    submitter : AbstractSubmitter
        The implemented submitter class

    Raises
    ------
    NotImplementedError
        If the name is not a supported submitter class
    """
    implemented = ("local",)

    if name == "local":
        submitter = LocalSubmitter(*args, **kwargs)
    else:
        msg = f"{name} is not a valid submitter class, choose " f"from {implemented}"
        raise NotImplementedError(msg)

    logging.debug("%s submitter selected", name)

    return submitter



"""Module containing the MakeProject class."""


import logging
from pathlib import Path
from typing import Optional

from bout_runners.submitter.local_submitter import LocalSubmitter
from bout_runners.utils.file_operations import get_caller_dir
from bout_runners.utils.names import get_exec_name, get_makefile_path


class MakeError(Exception):
    """Error class indicating that this is a Make error."""


class Make:
    """
    Class for making the project.

    Attributes
    ----------
    makefile_root_path : Path
        The path to the Makefile
    makefile_name : str
        The name of the Makefile
    makefile_path : Path
        Path to the makefile
    exec_name : str
        The name of the executable

    Methods
    -------
    run_make(force=False)
        Runs make in the self.makefile_root_path
    run_clean()
        Runs make clean in the self.makefile_root_path

    Examples
    --------
    >>> from bout_runners.make.make import Make
    ... from pathlib import Path
    ... path = Path('path', 'to', 'makefile_root_path')
    ... make_obj = Make(makefile_root_path=path)
    ... make_obj.run_make(force=True)
    """

    def __init__(
        self, makefile_root_path: Optional[Path] = None, makefile_name: None = None
    ) -> None:
        """
        Call the make file.

        Parameters
        ----------
        makefile_root_path : None or Path or str
            Root path of make file
            If None, the path of the root caller of MakeProject will be called
        makefile_name : None or str
            If set to None, it tries the following names, in order:
            'GNUmakefile', 'makefile' and 'Makefile'
        """
        if makefile_root_path is None:
            makefile_root_path = get_caller_dir()
        self.makefile_root_path = Path(makefile_root_path)
        logging.debug("self.makefile_root_path set to %s", makefile_root_path)

        self.makefile_name = makefile_name

        self.makefile_path = get_makefile_path(
            self.makefile_root_path, self.makefile_name
        )
        self.exec_name = get_exec_name(self.makefile_path)
        self.submitter = LocalSubmitter(self.makefile_root_path)

    def run_make(self, force: bool = False) -> None:
        """
        Execute the makefile.

        If an executable is found, nothing will be done unless 'force' is set to True

        Parameters
        ----------
        force : bool
            If True, make clean will be called prior to make
        """
        # If force: Run clean so that `made` returns false
        if force:
            self.run_clean()

        # Check if already made
        made = self.makefile_root_path.joinpath(self.exec_name).is_file()

        # Do nothing if already made
        if not made:
            make_str = (
                "make"
                if self.makefile_name is None
                else f"make -f {self.makefile_name}"
            )

            logging.info("Making the program")
            command = f"{make_str}"
            self.submitter.submit_command(command)

    def run_clean(self) -> None:
        """Run make clean."""
        make_str = (
            "make" if self.makefile_name is None else f"make -f {self.makefile_name}"
        )

        logging.info("Running make clean")
        command = f"{make_str} clean"
        self.submitter.submit_command(command)



"""Module containing the MetadataReader class."""


import re
from typing import Callable, Dict, List, Optional, Sequence, Tuple

from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_reader import DatabaseReader
from pandas import DataFrame


def drop_ids(func: Callable) -> Callable:
    """
    Return a function which remove excessive ids.

    Parameters
    ----------
    func : function
        A function returning a DataFrame

    Returns
    -------
    drop : function
        The function dropping the ids
    """

    def drop(self, *args, **kwargs) -> DataFrame:
        """
        Drop columns inplace.

        Parameters
        ----------
        self : object
            Self reference to the instance the function is belonging to
            Must contain self.drop_id
        args : tuple
            Arguments belonging to the input function
        kwargs : dict
            Keyword arguments to the input function

        Returns
        -------
        data_frame : DataFrame
            The DataFrame where the ids has been dropped
        """
        data_frame = func(self, *args, **kwargs)
        columns = tuple(data_frame.columns)
        drop_columns = list()

        if self.drop_id == "parameters":
            drop_columns = [col for col in columns if col.startswith("parameters.")]

        elif self.drop_id == "keep_run_id":
            drop_columns = [
                col
                for col in columns
                if (col.endswith(".id") and not col == "run.id") or col.endswith("_id")
            ]

        elif self.drop_id == "all_id":
            drop_columns = [
                col for col in columns if col.endswith(".id") or col.endswith("_id")
            ]

        if self.drop_id is not None:
            data_frame.drop(drop_columns, axis=1, inplace=True)

        return data_frame

    return drop


class MetadataReader:
    r"""
    Class for reading the metadata from the database.

    Attributes
    ----------
    __db_reader : DatabaseConnector
        The connection to the database
    __table_names : tuple
        Getter variable for table_names
    __table_column_dict : dict of tuple
        Getter variable for table_column_dict
    __table_connections : dict of tuple
        Getter variable for table_connections
    __sorted_columns : tuple
        Getter variable for sorted_columns
    table_names : tuple
         A tuple containing all names of the tables
    table_column_dict : dict of tuple
        A dict where the keys are table names, and the values are corresponding
        column names
    table_connections : dict of tuple
        A dict where the keys are tables, and the values are tuples of tables
        connected to the key table
    sorted_columns : tuple
        A tuple of the column names as they will be sorted in the all_metadata DataFrame
    date_columns : tuple
        Columns containing dates
    drop_id : None or str
        Specifies what id columns should be dropped when obtaining the metadata

    Methods
    -------
    get_all_metadata()
        Return all of the run metadata
    get_parameters_metadata()
        Return only the parameter part of the run metadata
    get_join_query(from_statement, columns, alias_columns, table_connections)
        Return the query string of a `SELECT` query with `INNER JOIN`
    __get_parameters_query()
        Return the parameters query string
    __get_sorted_columns()
        Return all columns sorted
    __get_table_connections()
        Return a dict containing the table connections
    __get_all_table_names()
        Return all the table names in the schema
    __get_table_column_dict()
        Return all the column names of the specified tables

    Examples
    --------
    >>> from bout_runners.database.database_connector import DatabaseConnector
    >>> db_connector = DatabaseConnector('test')
    >>> metadata_reader = MetadataReader(db_connector)
    >>> metadata_reader.get_parameters_metadata()
       bar.id  bar.foo  ... parameters.baz_id  parameters.foo_id
    0       1        1  ...                 1                  1
    1       2       10  ...                 1                  2
    2       2       10  ...                 1                  1

    [3 rows x 16 columns]

    >>> metadata_reader.get_all_metadata()
       run.id  ...                  system_info.version
    0       1  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    1       2  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    2       3  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    3       4  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    4       5  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    5       6  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    6       7  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019

    [7 rows x 43 columns]

    >>> metadata_reader.drop_id = 'all_id'
    >>> metadata_reader.get_all_metadata()
      run.latest_status  ...                  system_info.version
    0          complete  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    1          complete  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    2          complete  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    3          complete  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    4             error  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    5           running  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019
    6         submitted  ...  #1 SMP Thu Oct 17 19:31:58 UTC 2019

    [7 rows x 28 columns]
    """

    date_columns = (
        "run.start_time",
        "run.stop_time",
        "run.submitted_time",
        "file_modification.bout_lib_modified",
        "file_modification.project_executable_modified",
        "file_modification.project_makefile_modified",
    )

    def __init__(
        self,
        db_connector: Optional[DatabaseConnector] = None,
        drop_id: Optional[str] = "keep_run_id",
    ) -> None:
        """
        Set the database to use.

        Parameters
        ----------
        db_connector : DatabaseConnector or None
            The connection to the database
            If None: Default database connector will be used
        drop_id : None or str
            Specifies what id columns should be dropped when obtaining the metadata
            - None : No columns will be dropped
            - 'parameters' : All columns containing parameters ids
              will be dropped
            - 'keep_run_id' : Only the run.id of the id columns will be
              kept
            - 'all_id' : All id columns will be removed
        """
        self.drop_id = drop_id

        db_connector = db_connector if db_connector is not None else DatabaseConnector()
        self.__db_reader = DatabaseReader(db_connector)

        self.__table_names = self.__get_all_table_names()
        self.__table_column_dict = self.__get_table_column_dict()
        self.__table_connections = self.__get_table_connections()
        self.__sorted_columns = self.__get_sorted_columns()

        parameters_connections = {"parameters": self.__table_connections["parameters"]}
        parameters_tables = ("parameters", *parameters_connections["parameters"])
        self.__parameters_columns = tuple(
            str(col)
            for col in self.__sorted_columns
            if col.split(".")[0] in parameters_tables
        )

    @property
    def table_names(self) -> Tuple[str, ...]:
        """
        Set the properties of self.table_names.

        Returns
        -------
        self.__table_names : tuple
            A tuple containing all names of the tables
        """
        return self.__table_names

    @property
    def table_column_dict(self,) -> Dict[str, Tuple[str, ...]]:
        """
        Set the properties of self.table_column_dict.

        Returns
        -------
        self.__table_column_dict : dict of tuple
            A dict where the keys are table names, and the values are corresponding
            column names
        """
        return self.__table_column_dict

    @property
    def table_connection(self,) -> Dict[str, Tuple[str, ...]]:
        """
        Set the properties of self.table_connections.

        Returns
        -------
        self.__table_connections : dict of tuple
            A dict where the keys are tables, and the values are tuples of tables
            connected to the key table
        """
        return self.__table_connections

    @property
    def sorted_columns(self) -> Tuple[str, ...]:
        """
        Set the properties of self.sorted_columns.

        Returns
        -------
        self.__sorted_columns : tuple
            A tuple of the column names as they will be sorted in the all_metadata
            DataFrame
        """
        return self.__sorted_columns

    @drop_ids
    def get_all_metadata(self):
        """
        Return all of the run metadata.

        Returns
        -------
        DataFrame
            The DataFrame of the run metadata
        """
        parameters_query = self.__get_parameters_query()

        # Adding spaces and parenthesis
        parameter_sub_query = "\n".join(
            [f'{" " * 6}{line}' for line in parameters_query.split("\n")]
        )
        parameter_sub_query = (
            f"{parameter_sub_query[:5]}({parameter_sub_query[6:-1]}) " f"AS subquery"
        )

        # NOTE: The subquery names are the names of the columns after
        #       the query. We would like to rename them to
        #       sorted_columns. Hence the `columns` field and
        #       `alias_columns` field appears swapped
        subquery_columns = [
            f'subquery."{col}"' if col in self.__parameters_columns else col
            for col in self.sorted_columns
        ]
        # Remove the parameters from the table_connection to avoid
        # double joining
        table_connections = self.__table_connections.copy()
        table_connections.pop("parameters")
        unfinished_all_metadata_query = self.get_join_query(
            "run", subquery_columns, self.sorted_columns, table_connections
        )

        # Update the parameters columns
        all_metadata_query = unfinished_all_metadata_query.replace(
            " parameters ", f"\n{parameter_sub_query}\n"
        ).replace("= parameters.id", '= subquery."parameters.id"')

        return self.__db_reader.query(all_metadata_query, parse_dates=self.date_columns)

    @drop_ids
    def get_parameters_metadata(self):
        """
        Return only the parameter part of the run metadata.

        Returns
        -------
        DataFrame
            The DataFrame of the parameter metadata
        """
        parameters_query = self.__get_parameters_query()

        return self.__db_reader.query(parameters_query)

    @staticmethod
    def get_join_query(
        from_statement: str,
        columns: Sequence[str],
        alias_columns: Sequence[str],
        table_connections: Dict[str, Tuple[str, ...]],
    ) -> str:
        """
        Return the query string of a `SELECT` query with `INNER JOIN`.

        Notes
        -----
        The tables in `table_connection` is assumed to be joined by `id`s. I.e.
        `table_a` is connected to `table_b` by `table_b` having a column named
        `table_a_id` which corresponds to the `id` column of `table_a`

        Parameters
        ----------
        from_statement : str
            The statement after the `FROM` keyword in the query
            I.e.

            >>> f'SELECT * FROM {from_statement}'

        columns : array_like
            The columns to select from the tables
            I.e.

            >>> f'SELECT {columns} FROM *'

        alias_columns : array_like
            The name of the columns in the resulting table
            I.e.

            >>> f'SELECT {columns[0]} AS {alias_columns[0]} FROM *'

        table_connections : dict
            A dict where the keys are the table names, and the values are tuples
            containing table names connected to the key table as described in the
            note above

        Returns
        -------
        query : str
            The SQL-string which can be used to query where table in databases are
            joined through `INNER JOIN` operations
        """
        query = "SELECT\n"
        for column, alias in zip(columns, alias_columns):
            query += f'{" " * 7}{column} AS "{alias}",\n'
        # Remove last comma
        query = f"{query[:-2]}\n"
        query += f"FROM {from_statement}\n"
        for left_table in table_connections.keys():
            for right_table in table_connections[left_table]:
                query += (
                    f'{" " * 4}INNER JOIN {right_table} ON '
                    f"{left_table}."
                    f"{right_table}_id = {right_table}.id\n"
                )
        return query

    def __get_parameters_query(self) -> str:
        """
        Return the parameters query string.

        Returns
        -------
        parameters_query : str
            The SQL-string which can be used to query where table in databases are
            joined through `INNER JOIN` operations
        """
        parameter_connections = {"parameters": self.__table_connections["parameters"]}
        parameters_query = self.get_join_query(
            "parameters",
            self.__parameters_columns,
            self.__parameters_columns,
            parameter_connections,
        )
        return parameters_query

    def __get_sorted_columns(self) -> Tuple[str, ...]:
        """
        Return all columns sorted.

        The columns will be sorted alphabetically first by table name,
        then alphabetically by column name, with the following exceptions:

        1. The columns from the run table is presented first
        2. The id column is the first column in the table

        Returns
        -------
        tuple
            Dict containing the column names
            On the form

            >>> ('run.id',
            ...  'run.column_name_1',
            ...  'run.column_name_2',
            ...  ...
            ...  'table_name_1.column_name_1',
            ...  'table_name_1.column_name_2', ...)
        """
        sorted_columns: List[str] = list()
        table_names = sorted(self.table_column_dict.keys())
        table_names.pop(table_names.index("run"))
        table_names.insert(0, "run")
        for table_name in table_names:
            table_columns = list()
            for column_name in sorted(self.table_column_dict[table_name]):
                table_columns.append(f"{table_name}.{column_name}")
            table_columns.pop(table_columns.index(f"{table_name}.id"))
            table_columns.insert(0, f"{table_name}.id")
            sorted_columns = [*sorted_columns, *table_columns]
        return tuple(sorted_columns)

    def __get_table_connections(self) -> Dict[str, Tuple[str, ...]]:
        """
        Return a dict containing the table connections.

        Returns
        -------
        table_connection_dict : dict
            A dict telling which tables are connected to each other, where the key
            is the table under consideration and the value is a tuple containing the
            tables which have a key connection to the table under consideration
            On the form

            >>> {'table_1': ('table_2', 'table_3'),
            ...  'table_4': ('table_5',), ...}

        Raises
        ------
        RuntimeError
            If match is None
        """
        table_connection_dict = dict()
        pattern = re.compile("(.*)_id")

        for table, columns in self.table_column_dict.items():
            ids: List[str] = list()
            for column in columns:
                if "_id" in column:
                    match = pattern.match(column)
                    if match is None:
                        raise RuntimeError("match is None")
                    ids.append(match[1])
            if len(ids) > 0:
                table_connection_dict[table] = tuple(ids)

        return table_connection_dict

    def __get_all_table_names(self) -> Tuple[str, ...]:
        """
        Return all the table names in the schema.

        Returns
        -------
        tuple
            A tuple containing all names of the tables
        """
        query = (
            "SELECT name FROM sqlite_master\n"
            "WHERE\n"
            "    type ='table' AND\n"
            "    name NOT LIKE 'sqlite_%'"
        )
        # pylint: disable=no-member
        return tuple(self.__db_reader.query(query).loc[:, "name"])

    def __get_table_column_dict(self) -> Dict[str, Tuple[str, ...]]:
        """
        Return all the column names of the specified tables.

        Returns
        -------
        table_column_dict : dict of tuple
            Dict containing the column names
            On the form

            >>> {'table_1': ('table_1_column_1', ...),
            ...  'table_2': ('table_2_column_1', ...), ...}
        """
        table_column_dict = dict()

        query = "SELECT name FROM pragma_table_info('{}')"

        for table_name in self.table_names:
            # pylint: disable=no-member
            table_column_dict[table_name] = tuple(
                self.__db_reader.query(query.format(table_name)).loc[:, "name"]
            )

        return table_column_dict



"""Contains integration test for the runner."""

import contextlib
import os
from pathlib import Path
from typing import Callable, Iterator

from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_reader import DatabaseReader
from bout_runners.executor.bout_paths import BoutPaths
from bout_runners.executor.executor import Executor
from bout_runners.parameters.default_parameters import DefaultParameters
from bout_runners.parameters.final_parameters import FinalParameters
from bout_runners.parameters.run_parameters import RunParameters
from bout_runners.runner.bout_runner import BoutRunner
from bout_runners.submitter.local_submitter import LocalSubmitter


@contextlib.contextmanager
def change_directory(new_path: Path) -> Iterator[None]:
    """
    Change working directory and return to previous directory on exit.

    Parameters
    ----------
    new_path : Path
        Path to change to

    Yields
    ------
    None
        The function will revert to original directory on exit

    References
    ----------
    [1] https://stackoverflow.com/a/42441759/2786884
    [2] https://stackoverflow.com/a/13197763/2786884
    """
    previous_path = Path.cwd().absolute()
    os.chdir(str(new_path))
    try:
        yield
    finally:
        os.chdir(str(previous_path))


def assert_first_run(
    bout_paths: BoutPaths, db_connection: DatabaseConnector
) -> DatabaseReader:
    """
    Assert that the first run went well.

    Parameters
    ----------
    bout_paths : BoutPaths
        The object containing the paths
    db_connection : DatabaseConnector
        The database connection

    Returns
    -------
    db_reader : DatabaseReader
        The database reader object
    """
    db_reader = DatabaseReader(db_connection)
    assert bout_paths.bout_inp_dst_dir.joinpath("BOUT.dmp.0.nc").is_file()
    assert db_reader.check_tables_created()
    return db_reader


def assert_tables_has_len_1(
    db_reader: DatabaseReader, yield_number_of_rows_for_all_tables: Callable
) -> None:
    """
    Assert that tables has length 1.

    Parameters
    ----------
    db_reader : DatabaseReader
        The database reader object
    yield_number_of_rows_for_all_tables : function
        Function which returns the number of rows for all tables in a
        schema
    """
    number_of_rows_dict = yield_number_of_rows_for_all_tables(db_reader)
    assert sum(number_of_rows_dict.values()) == len(number_of_rows_dict.keys())


def assert_force_run(
    db_reader: DatabaseReader, yield_number_of_rows_for_all_tables: Callable
) -> None:
    """
    Assert that the force run is effective.

    Parameters
    ----------
    db_reader : DatabaseReader
        The database reader object
    yield_number_of_rows_for_all_tables : function
        Function which returns the number of rows for all tables in a
        schema
    """
    number_of_rows_dict = yield_number_of_rows_for_all_tables(db_reader)
    tables_with_2 = dict()
    tables_with_2["run"] = number_of_rows_dict.pop("run")
    # Assert that all the values are 1
    assert sum(number_of_rows_dict.values()) == len(number_of_rows_dict.keys())
    # Assert that all the values are 2
    assert sum(tables_with_2.values()) == 2 * len(tables_with_2.keys())


def test_bout_runners_from_directory(
    make_project: Path,
    yield_number_of_rows_for_all_tables: Callable,
    clean_default_db_dir: Path,
) -> None:
    """
    Test that the minimal BoutRunners setup works.

    This test will test that:
    1. We can execute a run from the (mocked) current work directory
    2. The metadata is properly stored
    3. We cannot execute the run again...
    4. ...unless we set force=True

    Parameters
    ----------
    make_project : Path
        The path to the conduction example
    yield_number_of_rows_for_all_tables : function
        Function which returns the number of rows for all tables in a schema
    clean_default_db_dir : Path
        Path to the default database directory
    """
    _ = clean_default_db_dir
    # Make project to save time
    project_path = make_project
    with change_directory(project_path):
        runner = BoutRunner()
        runner.run()

        bout_paths = runner.executor.bout_paths
        db_connection = runner.db_connector
    # Assert that the run went well
    db_reader = assert_first_run(bout_paths, db_connection)
    # Assert that all the values are 1
    assert_tables_has_len_1(db_reader, yield_number_of_rows_for_all_tables)

    # Check that the run will not be executed again
    with change_directory(project_path):
        runner.run()
    # Assert that all the values are 1
    assert_tables_has_len_1(db_reader, yield_number_of_rows_for_all_tables)

    # Check that force overrides the behaviour
    with change_directory(project_path):
        runner.run(force=True)
    assert_force_run(db_reader, yield_number_of_rows_for_all_tables)


def test_full_bout_runner(
    make_project: Path,
    yield_number_of_rows_for_all_tables: Callable,
    clean_default_db_dir: Path,
) -> None:
    """
    Test that the BoutRunner can execute a run.

    This test will test that:
    1. We can execute a run
    2. The metadata is properly stored
    3. We cannot execute the run again...
    4. ...unless we set force=True

    Parameters
    ----------
    make_project : Path
        The path to the conduction example
    yield_number_of_rows_for_all_tables : function
        Function which returns the number of rows for all tables in a schema
    clean_default_db_dir : Path
        Path to the default database directory
    """
    _ = clean_default_db_dir
    name = "test_bout_runner_integration"
    # Make project to save time
    project_path = make_project

    # Create the `bout_paths` object
    bout_paths = BoutPaths(
        project_path=project_path,
        bout_inp_src_dir=project_path.joinpath("data"),
        bout_inp_dst_dir=project_path.joinpath(name),
    )

    # Create the input objects
    run_parameters = RunParameters({"global": {"nout": 0}})
    default_parameters = DefaultParameters(bout_paths)
    final_parameters = FinalParameters(default_parameters, run_parameters)
    submitter = LocalSubmitter(bout_paths.project_path)
    executor = Executor(
        bout_paths=bout_paths, submitter=submitter, run_parameters=run_parameters,
    )
    db_connection = DatabaseConnector(name)

    # Run the project
    runner = BoutRunner(executor, db_connection, final_parameters)
    runner.run()

    # Assert that the run went well
    db_reader = assert_first_run(bout_paths, db_connection)
    # Assert that all the values are 1
    assert_tables_has_len_1(db_reader, yield_number_of_rows_for_all_tables)

    # Check that the run will not be executed again
    runner.run()
    # Assert that all the values are 1
    assert_tables_has_len_1(db_reader, yield_number_of_rows_for_all_tables)

    # Check that force overrides the behaviour
    runner.run(force=True)
    assert_force_run(db_reader, yield_number_of_rows_for_all_tables)



"""Module containing functions to extract names."""

from pathlib import Path
from typing import Optional

from bout_runners.make.read_makefile import (
    BoutMakefileVariableReader,
    MakefileReaderError,
)


def get_exec_name(makefile_path: Path) -> str:
    """
    Return the name of the project executable.

    This method first searches for the 'TARGET' variable in the makefile. If not
    found it infers the name from the 'SOURCEC' variable.

    Parameters
    ----------
    makefile_path : Path or str
        Path to the make file

    Returns
    -------
    exec_name : str
        Name of the executable
    """
    try:
        exec_name = BoutMakefileVariableReader(makefile_path, "TARGET").value
    except MakefileReaderError:
        exec_name = BoutMakefileVariableReader(makefile_path, "SOURCEC").value
        # Strip the name from the last .c*
        split_by = ".c"
        split_list = exec_name.split(split_by)
        split_to_join = split_list if len(split_list) == 1 else split_list[:-1]
        exec_name = f"{split_by}".join(split_to_join)

    return exec_name


def get_makefile_path(makefile_root_path: Path, makefile_name: Optional[str]) -> Path:
    """
    Return the makefile path.

    Parameters
    ----------
    makefile_root_path : Path or str
        Root path of make file
    makefile_name : None or str
        The name of the makefile.
        If set to None, it tries the following names, in order:
        'GNUmakefile', 'makefile' and 'Makefile'

    Returns
    -------
    makefile_path : Path
        Path to the makefile
    """
    if makefile_name is None:
        makefile_name = get_makefile_name(makefile_root_path)
    makefile_path = Path(makefile_root_path).joinpath(makefile_name)
    return makefile_path


def get_makefile_name(makefile_root_path: Path) -> str:
    """
    Search for a valid Makefile.

    The order of the search is 'GNUmakefile', 'makefile' and 'Makefile'

    Parameters
    ----------
    makefile_root_path : Path
        Path to the root directory of the Makefile

    Returns
    -------
    makefile_name : str
        The name of the Makefile

    Raises
    ------
    FileNotFoundError
        If none of the valid makefile names are found
    """
    possible_names = ("GNUmakefile", "Makefile", "makefile")

    makefile_name = None

    for name in possible_names:
        if makefile_root_path.joinpath(name).is_file():
            makefile_name = name
            break

    if makefile_name is None:
        msg = (
            f"Could not find a valid Makefile name in {makefile_root_path}. "
            f"Valid Makefile names are {' ,'.join(possible_names)}"
        )
        raise FileNotFoundError(msg)

    return makefile_name



"""Contains unittests for the metadata_recorder."""


from pathlib import Path
from typing import Callable

from bout_runners.metadata.metadata_recorder import MetadataRecorder
from bout_runners.parameters.default_parameters import DefaultParameters
from bout_runners.parameters.final_parameters import FinalParameters
from bout_runners.submitter.processor_split import ProcessorSplit


def test_metadata_recorder(
    yield_bout_path_conduction: Callable,
    get_default_parameters: DefaultParameters,
    make_project: Path,
    make_test_schema: Callable,
    yield_number_of_rows_for_all_tables: Callable,
) -> None:
    """
    Test the metadata recorder.

    Specifically this test will test that:
    1. It is possible to make a new entry
    2. The newly created schema only have 1 row from the created entry
    3. No new entries are made when trying to make the entry again
    4. All the tables still have one row
    5. Create a new entry after changing the run parameters
    6. Check that a new entry is created
    7. See that the `split` table and the `run` table have two rows, whereas the rest
       have one
    8. Check that it is possible to forcefully make an entry to the run table

    Parameters
    ----------
    yield_bout_path_conduction : function
        Function which makes the BoutPaths object for the conduction example
    get_default_parameters : DefaultParameters
        The DefaultParameters object
    make_project : Path
        The path to the conduction example
    make_test_schema : function
        The function making the schema (i.e. making all the tables)
    yield_number_of_rows_for_all_tables : function
        Function which returns the number of rows for all tables in a schema
    """
    # NOTE: If the project is not made, the metadata recorder will
    # fail when the get_file_modification is trying to get the last
    # edited time of the executable
    _ = make_project
    db_connection, _ = make_test_schema("test_metadata_recorder")
    bout_paths = yield_bout_path_conduction("test_metadata_recorder")
    default_parameters = get_default_parameters
    final_parameters = FinalParameters(default_parameters)

    metadata_recorder = MetadataRecorder(db_connection, bout_paths, final_parameters)

    # Assert that this is a new entry
    run_id = metadata_recorder.capture_new_data_from_run(ProcessorSplit())
    assert run_id is None
    # Assert that all the values are 1
    number_of_rows_dict = yield_number_of_rows_for_all_tables(
        metadata_recorder.db_reader
    )
    assert sum(number_of_rows_dict.values()) == len(number_of_rows_dict.keys())

    # Assert that this is not a new entry
    run_id = metadata_recorder.capture_new_data_from_run(ProcessorSplit())
    assert run_id == 1
    # Assert that all the values are 1
    number_of_rows_dict = yield_number_of_rows_for_all_tables(
        metadata_recorder.db_reader
    )
    assert sum(number_of_rows_dict.values()) == len(number_of_rows_dict.keys())

    # Create a new entry in the split table
    run_id = metadata_recorder.capture_new_data_from_run(
        ProcessorSplit(number_of_nodes=2)
    )
    # Assert that a new entry has been made (the number of rows in
    # the tables will be counted when checking the forceful entry)
    assert run_id is None

    # Forcefully make an entry
    run_id = metadata_recorder.capture_new_data_from_run(
        ProcessorSplit(number_of_nodes=2), force=True
    )
    # Assert that a new entry has been made
    assert run_id == 2
    number_of_rows_dict = yield_number_of_rows_for_all_tables(
        metadata_recorder.db_reader
    )
    tables_with_2 = dict()
    tables_with_2["split"] = number_of_rows_dict.pop("split")
    tables_with_3 = dict()
    tables_with_3["run"] = number_of_rows_dict.pop("run")
    # Assert that all the values are 1
    assert sum(number_of_rows_dict.values()) == len(number_of_rows_dict.keys())
    # Assert that all the values are 2
    assert sum(tables_with_2.values()) == 2 * len(tables_with_2.keys())
    # Assert that all the values are 3
    assert sum(tables_with_3.values()) == 3 * len(tables_with_3.keys())



"""Contains unittests for the database reader."""


from typing import Callable

from bout_runners.database.database_reader import DatabaseReader


def test_db_reader(make_test_database: Callable, write_to_split: Callable) -> None:
    """
    Test we can create read from the database.

    Specifically this test that:
    1. We can make a query
    2. That an empty db has not been populated
    3. That a populated db has table entries
    4. Check that we can extract the id for a given set of values which exist
    5. Check that no id is returned if a given set of values is not found in the
       database
    6. That we can get the latest row id which has been written to

    Parameters
    ----------
    make_test_database : function
        Function which returns the database connection
    write_to_split : function
        Function returning the database connection where `split` has been populated
    """
    empty_db_connection = make_test_database("empty_read_test")
    empty_db_reader = DatabaseReader(empty_db_connection)

    # Check that we can make a query
    table = empty_db_reader.query("SELECT 1+1 AS col")
    assert table.loc[0, "col"] == 2  # pylint: disable=no-member

    # Check that the tables has not been created in an empty db
    assert not empty_db_reader.check_tables_created()

    db_connection = write_to_split("read_test")
    db_reader = DatabaseReader(db_connection)

    # Check that tables exist
    assert db_reader.check_tables_created()

    # As write_to_split writes to the split table, we can get the
    # written values with the following query
    table = db_reader.query("SELECT * FROM split")
    entries_dict = table.to_dict(orient="records")[0]

    # Remove the 'id'
    entries_dict.pop("id")

    row_id = db_reader.get_entry_id("split", entries_dict)
    assert row_id == 1

    # Modify entries_dict so that row_id returns None
    entries_dict[list(entries_dict.keys())[0]] += 1
    new_row_id = db_reader.get_entry_id("split", entries_dict)
    assert new_row_id is None

    # Assert that get_latest_row_id is working
    assert db_reader.get_latest_row_id() == 1



"""Contains the abstract submitter class."""


from abc import ABC, abstractmethod
from typing import Any


class AbstractSubmitter(ABC):
    """The abstract base class of the submitters."""

    @abstractmethod
    def submit_command(self, command: str):
        """
        Submit a command.

        Parameters
        ----------
        command : str
            Command to submit
        """

    @property
    @abstractmethod
    def pid(self):
        """Return the process id."""

    @abstractmethod
    def _raise_submit_error(self, result: Any):
        """
        Raise error if submission failed.

        Parameters
        ----------
        result : object
            The result from the subprocess
        """



"""Contains the class setting the BOUT++ paths."""


import logging
import shutil
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from bout_runners.utils.file_operations import get_caller_dir


class BoutPaths:
    """
    Class which sets the paths.

    Attributes
    ----------
    __project_path : None or Path
        Getter and setter variable for project_path
    __bout_inp_src_dir : None or Path
        Getter and setter variable for bout_inp_src_dir
    __bout_inp_dst_dir : None or Path
        Getter and setter variable for bout_inp_dst_dir
    project_path : Path
        The root path of the project
    bout_inp_src_dir : Path
        The path to the BOUT.inp source directory
    bout_inp_dst_dir : Path
        The path to the BOUT.inp destination directory

    Methods
    -------
    _copy_inp()
        Copy BOUT.inp from bout_inp_src_dir to bout_inp_dst_dir

    Examples
    --------
    >>> bout_paths = BoutPaths()
    >>> bout_paths.project_path
    Path(/root/BOUT-dev/examples/conduction)

    >>> bout_paths.bout_inp_src_dir
    Path(/root/BOUT-dev/examples/conduction/data)

    >>> bout_paths.bout_inp_dst_dir
    Path(/root/BOUT-dev/examples/conduction/2020-02-12_21-59-00_227295)

    >>> bout_paths.bout_inp_dst_dir = 'foo'
    Path(/root/BOUT-dev/examples/conduction/foo)
    """

    def __init__(
        self,
        project_path: Optional[Union[Path, str]] = None,
        bout_inp_src_dir: Optional[Union[Path, str]] = None,
        bout_inp_dst_dir: Optional[Union[Path, str]] = None,
    ) -> None:
        """
        Set the paths.

        Parameters
        ----------
        project_path : None or Path or str
            Root path of make file
            If None, the path of the path of the root caller will be used
        bout_inp_src_dir : None or str or Path
            The path to the BOUT.inp source directory (relative to self.project_path)
            If None, data will be used
        bout_inp_dst_dir : None or str or Path
            The path to the BOUT.inp bout_inp_dst_dir directory (relative to
            self.project_path)
            If None, the current time will be used
        """
        # Declare variables to be used in the getters and setters
        self.__project_path: Optional[Path] = None
        self.__bout_inp_src_dir: Optional[Path] = None
        self.__bout_inp_dst_dir: Optional[Path] = None

        # NOTE: type: ignore due to https://github.com/python/mypy/issues/3004
        # Set the project path
        self.project_path = project_path  # type: ignore

        # Set the bout_inp_src_dir
        self.bout_inp_src_dir = bout_inp_src_dir  # type: ignore

        # Set the bout_inp_dst_dir
        self.bout_inp_dst_dir = bout_inp_dst_dir  # type: ignore

    @property
    def project_path(self) -> Path:
        """
        Set the properties of self.project_path.

        If None is specified, the path of the path of the root caller will be used

        Returns
        -------
        Path
            Absolute path to the root of make file
        """
        # NOTE: The setter guarantees that self.__project_path is not None, however
        #       mypy complains if just self.__project_path is returned
        #       This may be a side effect of https://github.com/python/mypy/issues/3004
        return Path(self.__project_path) if self.__project_path is not None else Path()

    @project_path.setter
    def project_path(self, project_path: Optional[Union[Path, str]]) -> None:
        if project_path is None:
            project_path = get_caller_dir()
        project_path = Path(project_path).absolute()
        self.__project_path = project_path
        logging.debug("self.project_path set to %s", project_path)

    @property
    def bout_inp_src_dir(self) -> Path:
        """
        Set the properties of bout_inp_src_dir.

        The setter will convert bout_inp_src_dir an absolute path (as the input is
        relative to the project path), check that the directory exists, and copy the
        BOUT.inp file to the bout_inp_dst_dir path (self.bout_inp_dst_dir)

        The input should be the path to the BOUT.inp source directory (relative to
        self.project_path)
        If None, "data" will be used

        Returns
        -------
        self.__bout_inp_src_dir : Path
            The absolute path to the BOUT.inp source directory

        Raises
        ------
        FileNotFoundError
            If no BOUT.inp file is found in the directory
        """
        # NOTE: The setter guarantees that self.__project_path is not None, however
        #       mypy complains if just self.__project_path is returned
        #       This may be a side effect of https://github.com/python/mypy/issues/3004
        return (
            Path(self.__bout_inp_src_dir)
            if self.__bout_inp_src_dir is not None
            else Path()
        )

    @bout_inp_src_dir.setter
    def bout_inp_src_dir(self, bout_inp_src_dir: Optional[Union[Path, str]]) -> None:
        bout_inp_src_dir = (
            Path(bout_inp_src_dir) if bout_inp_src_dir is not None else Path("data")
        )

        self.__bout_inp_src_dir = self.project_path.joinpath(bout_inp_src_dir)

        if not self.__bout_inp_src_dir.joinpath("BOUT.inp").is_file():
            raise FileNotFoundError(
                f"No BOUT.inp file found in " f"{self.__bout_inp_src_dir}"
            )

        logging.debug("self.bout_inp_src_dir set to %s", self.__bout_inp_src_dir)

        # Copy file to bout_inp_dst_dir if set
        if self.bout_inp_dst_dir is not None:
            self._copy_inp()

    @property
    def bout_inp_dst_dir(self) -> Path:
        """
        Set the properties of bout_inp_dst_dir.

        The setter will convert bout_inp_dst_dir an absolute path (as the input is
        relative to the project path), and copy BOUT.inp from self.bout_inp_src_dir
        to self.bout_inp_dst_dir

        The input should be the path to the BOUT.inp bout_inp_dst_dir directory
        (relative to self.project_path)
        If None, the current time will be used

        Returns
        -------
        Path
            Path to the destination directory
        """
        # NOTE: The setter guarantees that self.__project_path is not None, however
        #       mypy complains if just self.__project_path is returned
        #       This may be a side effect of https://github.com/python/mypy/issues/3004
        return (
            Path(self.__bout_inp_dst_dir)
            if self.__bout_inp_dst_dir is not None
            else Path()
        )

    @bout_inp_dst_dir.setter
    def bout_inp_dst_dir(self, bout_inp_dst_dir: Optional[Union[Path, str]]) -> None:
        time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S_%f")
        bout_inp_dst_dir = (
            Path(bout_inp_dst_dir) if bout_inp_dst_dir is not None else Path(time)
        )

        self.__bout_inp_dst_dir = self.project_path.joinpath(bout_inp_dst_dir)
        self.__bout_inp_dst_dir.mkdir(exist_ok=True, parents=True)
        logging.debug("self.bout_inp_dst_dir set to %s", self.__bout_inp_dst_dir)

        self._copy_inp()

    def _copy_inp(self) -> None:
        """Copy BOUT.inp from bout_inp_src_dir to bout_inp_dst_dir."""
        if self.bout_inp_src_dir != self.bout_inp_dst_dir:
            src = self.bout_inp_src_dir.joinpath("BOUT.inp")
            dst = self.bout_inp_dst_dir.joinpath(src.name)
            shutil.copy(src, dst)
            logging.debug("Copied %s to %s", src, dst)



"""Contains the executor class."""


from typing import Optional

from bout_runners.executor.bout_paths import BoutPaths
from bout_runners.make.make import Make
from bout_runners.parameters.run_parameters import RunParameters
from bout_runners.submitter.local_submitter import LocalSubmitter


class Executor:
    r"""
    Executes the command for submitting a run.

    Attributes
    ----------
    __bout_paths : BoutPaths
        Getter variable for project_path
    __run_parameters : RunParameters
        Object containing the run parameters
    __make : Make
        Object for making the project
    __command : str
        The terminal command for executing the run
    submitter : AbstractSubmitter
        Object containing the submitter
    bout_paths : BoutPaths
        Object containing the paths

    Methods
    -------
    get_execute_command()
        Return the execute command string
    execute()
        Execute a BOUT++ run

    Examples
    --------
    The easiest way to use the Executor is to run a script from the root directory of
    the project (i.e. where the `Makefile` and `data` directory are normally
    situated. The script can simply call

    >>> Executor().execute()

    and `Executor` takes care of the rest.

    A more elaborate example where all the dependency objects are built manually:

    Import the dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.submitter.local_submitter import LocalSubmitter

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Create the executor object

    >>> run_parameters = RunParameters({'global': {'nout': 0}})
    >>> executor = Executor(
    ...     bout_paths=bout_paths,
    ...     submitter=LocalSubmitter(bout_paths.project_path),
    ...     run_parameters=run_parameters)

    Execute the run

    >>> executor.execute()
    """

    def __init__(
        self,
        bout_paths: Optional[BoutPaths] = None,
        submitter: Optional[LocalSubmitter] = None,
        run_parameters: Optional[RunParameters] = None,
    ) -> None:
        """
        Set the input parameters.

        Parameters
        ----------
        bout_paths : BoutPaths or None
            Object containing the paths
            If None, default BoutPaths values will be used
        submitter : AbstractSubmitter
            Object containing the submitter
        run_parameters : RunParameters or None
            Object containing the run parameters
            If None, default parameters will be used
        """
        # Set member data
        # NOTE: We are not setting the default as a keyword argument
        #       as this would mess up the paths
        self.submitter = submitter if submitter is not None else LocalSubmitter()
        self.__bout_paths = bout_paths if bout_paths is not None else BoutPaths()
        self.__run_parameters = (
            run_parameters if run_parameters is not None else RunParameters()
        )
        self.__make = Make(self.__bout_paths.project_path)
        self.__command = self.get_execute_command()

    @property
    def bout_paths(self) -> BoutPaths:
        """
        Set the properties of self.bout_paths.

        Returns
        -------
        self.__bout_paths : BoutPaths
            Object containing the paths

        Notes
        -----
        The bout_paths is read only
        """
        return self.__bout_paths

    @property
    def run_parameters(self):
        """
        Set the properties of self.run_parameters.

        Returns
        -------
        self.__run_parameters : RunParameters
            Object containing the run parameters

        Notes
        -----
        The run_parameters is read only
        """
        return self.__run_parameters

    def get_execute_command(self) -> str:
        """
        Return the execute command string.

        Returns
        -------
        command : str
            The terminal command for executing the run
        """
        mpi_cmd = "mpirun -np"

        # NOTE: No spaces if parameters are None
        command = (
            f"{mpi_cmd} "
            f"{self.submitter.processor_split.number_of_processors} "
            f"./{self.__make.exec_name} "
            f"-d {self.__bout_paths.bout_inp_dst_dir} "
            f"{self.__run_parameters.run_parameters_str}"
        )
        return command

    def execute(self) -> None:
        """Execute a BOUT++ run."""
        # Make the project if not already made
        self.__make.run_make()
        # Submit the command
        self.submitter.submit_command(self.__command)



"""Module containing the MetadataUpdater class."""


from datetime import datetime
from typing import Union

from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_writer import DatabaseWriter


class MetadataUpdater:
    r"""
    Class which updates dynamic data about the run.

    Attributes
    ----------
    __db_writer : DatabaseWriter
        Class used to write to the database
    run_id : int
        The id of the run to update

    Methods
    -------
    update_start_time(start_time)
        Update the start time
    update_end_time(end_time)
        Update the end time
    update_latest_status(status)
        Update the latest status

    Examples
    --------
    >>> from pathlib import Path
    >>> from datetime import datetime
    >>> from bout_runners.database.database_connector import DatabaseConnector
    >>> db_root_path = Path().joinpath('path', 'to', 'db_root')
    >>> db_connector = DatabaseConnector('name_of_db', db_root_path)
    >>> run_id = 1  # This must be a valid id in the run table
    >>> metadata_updater = MetadataUpdater(db_connector, run_id)
    >>> metadata_updater.update_start_time(datetime.now())
    >>> metadata_updater.update_stop_time(datetime.now())
    >>> metadata_updater.update_latest_status('error')
    """

    def __init__(self, db_connector: DatabaseConnector, run_id: int) -> None:
        """
        Set the database and id to use.

        Parameters
        ----------
        db_connector : DatabaseConnector
            The database connector
        run_id : int
            The id of the run to update
        """
        self.__db_writer = DatabaseWriter(db_connector)
        self.run_id = run_id

    def update_start_time(self, start_time: datetime) -> None:
        """
        Update the start time.

        Parameters
        ----------
        start_time : datetime
            The start time of the execution
        """
        self.update_column("start_time", start_time)

    def update_stop_time(self, end_time: datetime) -> None:
        """
        Update the end time.

        Parameters
        ----------
        end_time : datetime
            The end time of the execution
        """
        self.update_column("stop_time", end_time)

    def update_latest_status(self, status: str) -> None:
        """
        Update the latest status.

        Parameters
        ----------
        status : str
            The latest status
        """
        self.update_column("latest_status", status)

    def update_column(self, column: str, value: Union[datetime, str]) -> None:
        """
        Update a field with a certain row in the run table.

        Parameters
        ----------
        column : str
            The column to update
        value : object
            The updating value
        """
        update_str = self.__db_writer.create_update_string(
            field_names=(column,),
            table_name="run",
            search_condition=f"id = {self.run_id}",
        )
        self.__db_writer.update(update_str, (value,))



"""Module containing the DatabaseReader class."""


from typing import Iterable, Mapping, Optional, Union

import pandas as pd
from bout_runners.database.database_connector import DatabaseConnector
from numpy import int64
from pandas import DataFrame


class DatabaseReader:
    r"""
    Class for reading the schema of the database.

    Attributes
    ----------
    db_connector : DatabaseConnector
        The database object to read from

    Methods
    -------
    query(query_str)
        Make a query to the database
    get_latest_row_id()
        Return the latest row id
    get_entry_id(table_name, entries_dict)
        Get the id of a table entry
    check_tables_created()
        Check if the tables is created in the database

    Examples
    --------
    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.parameters.default_parameters import DefaultParameters
    >>> from bout_runners.parameters.final_parameters import FinalParameters
    >>> from bout_runners.database.database_connector import DatabaseConnector
    >>> from bout_runners.database.database_creator import DatabaseCreator
    >>> from bout_runners.database.database_writer import DatabaseWriter

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Obtain the parameters

    >>> default_parameters = DefaultParameters(bout_paths)
    >>> final_parameters = FinalParameters(default_parameters)
    >>> final_parameters_dict = final_parameters.get_final_parameters()
    >>> final_parameters_as_sql_types = \
    ...     final_parameters.cast_to_sql_type(final_parameters_dict)

    Create the database

    >>> db_connection = DatabaseConnector('name')
    >>> db_creator = DatabaseCreator(db_connection)
    >>> db_creator.create_all_schema_tables(final_parameters_as_sql_types)

    Write to the database

    >>> db_writer = DatabaseWriter(db_connection)
    >>> dummy_split_dict = {'number_of_processors': 1,
    ...                     'number_of_nodes': 2,
    ...                     'processors_per_node': 3}
    >>> db_writer.create_entry('split', dummy_split_dict)

    Read from the database

    >>> db_reader = DatabaseReader(db_connection)
    >>> db_reader.query('SELECT * FROM split')
        id  number_of_processors  number_of_nodes  processors_per_node
     0   1                     1                1                    1

    >>> db_reader.get_latest_row_id()
    1

    >>> db_reader.get_entry_id('split', dummy_split_dict)
    1

    >>> db_reader.check_tables_created()
    True
    """

    def __init__(self, db_connector: DatabaseConnector) -> None:
        """
        Set the database to use.

        Parameters
        ----------
        db_connector : DatabaseConnector
            The database object to read from
        """
        self.db_connector = db_connector

    def query(
        self,
        query_str: str,
        params: Optional[Iterable[Union[str, float, int, bool, None]]] = None,
        **kwargs,
    ) -> DataFrame:
        """
        Make a query to the database.

        Parameters
        ----------
        query_str : str
            The query to execute
        params : array_like
            Array of parameters
            Used to protect against SQL injection

            >>> self.query("SELECT ? FROM table WHERE foo = ?", params=("bar", "baz"))

        kwargs : dict
            Additional keyword parameters to pd.read_sql_query

        Returns
        -------
        table : DataFrame
            The result of a query as a DataFrame
        """
        table = pd.read_sql_query(
            query_str, self.db_connector.connection, params=params, **kwargs
        )
        return table

    def get_latest_row_id(self) -> int64:
        """
        Return the latest row id.

        Returns
        -------
        row_id : int
            The latest row inserted row id
        """
        # https://stackoverflow.com/questions/3442033/sqlite-how-to-get-value-of-auto-increment-primary-key-after-insert-other-than
        # pylint: disable=no-member
        row_id = self.query("SELECT last_insert_rowid() AS id").loc[0, "id"]
        return row_id

    def get_entry_id(
        self, table_name: str, entries_dict: Mapping[str, Union[int, str, float, None]]
    ) -> Optional[int]:
        """
        Get the id of a table entry.

        Parameters
        ----------
        table_name : str
            Name of the table to check
        entries_dict : dict
            Dictionary containing the entries as key value pairs

        Returns
        -------
        row_id : int or None
            The id of the hit
            If none is found, None is returned
        """
        # NOTE: About checking for existence
        # https://stackoverflow.com/questions/9755860/valid-query-to-check-if-row-exists-in-sqlite3
        # NOTE: About SELECT 1
        # https://stackoverflow.com/questions/7039938/what-does-select-1-from-do
        where_statements_list = list()
        where_values = list()
        for field, val in entries_dict.items():
            val = f"{val}" if isinstance(val, str) else val
            where_statements_list.append(f'{" "*7}AND {field}=?')
            where_values.append(val)
        where_statements_list[0] = where_statements_list[0].replace("AND", "WHERE")
        where_statements_str = "\n".join(where_statements_list)

        # NOTE: Protection against SQL injection through the use of ? in for-loop above
        query_str = f"SELECT id\nFROM {table_name}\n{where_statements_str}"  # nosec

        table = self.query(query_str, params=where_values)
        # NOTE: We explicitly cast to int, as sqlite3 will cast
        #       np.int64 to bytes
        # pylint: disable=no-member
        row_id = None if table.empty else int(table.loc[0, "id"])

        return row_id

    def check_tables_created(self) -> bool:
        """
        Check if the tables is created in the database.

        Returns
        -------
        bool
            Whether or not the tables are created
        """
        query_str = 'SELECT name FROM sqlite_master WHERE type="table"'

        table = self.query(query_str)
        return len(table.index) != 0



"""Contains unittests for the MetadataUpdater."""


from datetime import datetime
from typing import Callable


def test_update_start_time(get_metadata_updater_and_db_reader: Callable) -> None:
    """
    Test if it's possible to update the start time.

    Parameters
    ----------
    get_metadata_updater_and_db_reader : function
        Function which returns the MetadataUpdater object with initialized with
        connection to the database and a corresponding DatabaseReader object
    """
    metadata_updater, db_reader = get_metadata_updater_and_db_reader("start_time")
    now = datetime.now()
    run_id = metadata_updater.run_id
    metadata_updater.update_start_time(now)
    result_df = db_reader.query(
        "SELECT start_time FROM run WHERE id = ?", params=(run_id,)
    )
    assert result_df.loc[0, "start_time"] == str(now)


def test_update_stop_time(get_metadata_updater_and_db_reader: Callable) -> None:
    """
    Test if it's possible to update the stop time.

    Parameters
    ----------
    get_metadata_updater_and_db_reader : function
        Function which returns the MetadataUpdater object with initialized with
        connection to the database and a corresponding DatabaseReader object
    """
    metadata_updater, db_reader = get_metadata_updater_and_db_reader("stop_time")
    now = datetime.now()
    run_id = metadata_updater.run_id
    metadata_updater.update_stop_time(now)
    result_df = db_reader.query(
        "SELECT stop_time FROM run WHERE id = ?", params=(run_id,)
    )
    assert result_df.loc[0, "stop_time"] == str(now)


def test_update_latest_status(get_metadata_updater_and_db_reader: Callable) -> None:
    """
    Test if it's possible to update the latest status.

    Parameters
    ----------
    get_metadata_updater_and_db_reader : function
        Function which returns the MetadataUpdater object with initialized with
        connection to the database and a corresponding DatabaseReader object
    """
    metadata_updater, db_reader = get_metadata_updater_and_db_reader("latest_status")
    latest_status = "foobar"
    run_id = metadata_updater.run_id
    metadata_updater.update_latest_status(latest_status)
    result_df = db_reader.query(
        "SELECT latest_status FROM run WHERE id = ?", params=(run_id,)
    )
    assert result_df.loc[0, "latest_status"] == latest_status



"""Module containing the DatabaseWriter class."""


import logging
import re
from typing import Any, Mapping, Sequence, Tuple, Union

from bout_runners.database.database_connector import DatabaseConnector


class DatabaseWriter:
    r"""
    Class for writing to the schema of the database.

    Attributes
    ----------
    db_connector : DatabaseConnector
        The database object to write to

    Methods
    -------
    create_insert_string(field_names, table_name)
        Create a question mark style string for database insertions
    insert(insert_str, values)
        Insert to the database
    create_entry(table_name, entries_dict)
        Create a database entry

    Examples
    --------
    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.parameters.default_parameters import DefaultParameters
    >>> from bout_runners.parameters.final_parameters import FinalParameters
    >>> from bout_runners.database.database_connector import DatabaseConnector
    >>> from bout_runners.database.database_creator import DatabaseCreator

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Obtain the parameters

    >>> default_parameters = DefaultParameters(bout_paths)
    >>> final_parameters = FinalParameters(default_parameters)
    >>> final_parameters_dict = final_parameters.get_final_parameters()
    >>> final_parameters_as_sql_types = \
    ...     final_parameters.cast_to_sql_type(final_parameters_dict)

    Create the database

    >>> db_connection = DatabaseConnector('name')
    >>> db_creator = DatabaseCreator(db_connection)
    >>> db_creator.create_all_schema_tables(
    ...     final_parameters_as_sql_types)

    Write to the database

    >>> db_writer = DatabaseWriter(db_connection)
    >>> dummy_split_dict = {'number_of_processors': 1,
    ...                     'number_of_nodes': 2,
    ...                     'processors_per_node': 3}
    >>> db_writer.create_entry('split', dummy_split_dict)
    """

    def __init__(self, db_connector: DatabaseConnector) -> None:
        """
        Set the database to use.

        Parameters
        ----------
        db_connector : DatabaseConnector
            The database object to write to
        """
        self.db_connector = db_connector

    @staticmethod
    def create_insert_string(field_names: Sequence[str], table_name: str) -> str:
        """
        Create a question mark style string for database insertions.

        Values must be provided separately in the execution statement

        Parameters
        ----------
        field_names : array_like
            Names of the fields to populate
        table_name : str
            Name of the table to use for the insertion

        Returns
        -------
        insert_str : str
            The string to be used for insertion
        """
        # From
        # https://stackoverflow.com/a/14108554/2786884
        columns = ", ".join(field_names)
        placeholders = ", ".join("?" * len(field_names))
        insert_str = (
            f"INSERT INTO {table_name} " f"({columns}) " f"VALUES ({placeholders})"
        )
        return insert_str

    @staticmethod
    def create_update_string(
        field_names: Tuple[str, ...], table_name: str, search_condition: str,
    ) -> str:
        """
        Create a question mark style string for database update.

        Values must be provided separately in the execution statement

        Parameters
        ----------
        field_names : array_like
            Names of the fields to populate
        table_name : str
            Name of the table to use for the update
        search_condition : str
            Condition for the update
            Example

            >>> 'id = 3 AND col = 42'

        Returns
        -------
        insert_str : str
            The string to be used for update
        """
        placeholders = ""
        for col in field_names:
            placeholders += f'{" " * 4}{col} = ?,\n'
        # Remove last comma
        placeholders = f"{placeholders[:-2]}\n"

        update_str = (
            f"UPDATE {table_name}\n" f"SET\n{placeholders}" f"WHERE {search_condition}"
        )
        return update_str

    def insert(self, insert_str: str, values: Any) -> None:
        """
        Insert to the database.

        Parameters
        ----------
        insert_str : str
            The write statement to execute
        values : tuple
            Values to be inserted in the query

        Raises
        ------
        ValueError
            If the insert_str is not understood
        """
        # Obtain the table name
        pattern = r"INSERT INTO (\w*)"
        match = re.match(pattern, insert_str)
        if match is None:
            raise ValueError(f'insert_str "{insert_str}" not understood')

        table_name = match.group(1)

        self.db_connector.execute_statement(insert_str, *values)

        logging.info("Made insertion to %s", table_name)

    def update(self, update_str: str, values: Any,) -> None:
        """
        Insert to the database.

        Parameters
        ----------
        update_str : str
            The update statement to execute
        values : tuple
            Values to be inserted in the query

        Raises
        ------
        ValueError
            If update_str is not understood
        """
        # Obtain the table name
        pattern = r"UPDATE (\w*)"
        match = re.match(pattern, update_str)

        if match is None:
            raise ValueError(f'update_str "{update_str}" not understood')
        table_name = match.group(1)

        pattern = r"WHERE (.*)"
        match = re.search(pattern, update_str)

        if match is None:
            raise ValueError(f'update_str "{update_str}" not understood')
        condition = match.group(1)

        self.db_connector.execute_statement(update_str, *values)

        logging.info("Updated table %s, where %s", table_name, condition)

    def create_entry(
        self, table_name: str, entries_dict: Mapping[str, Union[int, str, float, None]]
    ) -> None:
        """
        Create a database entry.

        Parameters
        ----------
        table_name : str
            Name of the table
        entries_dict : dict
            Dictionary containing the entries as key value pairs
        """
        keys = entries_dict.keys()
        values = tuple(entries_dict.values())
        insert_str = self.create_insert_string(tuple(keys), table_name)
        self.insert(insert_str, values)



"""Module containing database utils."""


import logging
import platform
from pathlib import Path

# NOTE: subprocess can be vulnerable if shell=True
#       However, CalledProcessError has no known security vulnerabilities
from subprocess import CalledProcessError  # nosec
from typing import Dict

from bout_runners.submitter.local_submitter import LocalSubmitter
from bout_runners.utils.file_operations import get_modified_time
from bout_runners.utils.paths import get_bout_directory


def get_system_info_as_sql_type() -> Dict[str, str]:
    """
    Return the SQL types of the system information.

    Returns
    -------
    sys_info_dict : dict
        Dictionary with the attributes of the system as keys and the type as values
    """
    attributes = get_system_info()

    sys_info_dict = {att: "TEXT" for att in attributes.keys()}

    return sys_info_dict


def get_file_modification(
    project_path: Path, makefile_path: Path, exec_name: str
) -> Dict[str, str]:
    """
    Return the file modification info.

    Parameters
    ----------
    project_path : Path
        Path to the project
    makefile_path : Path
            Path to the project makefile
    exec_name : str
        Name of the executable

    Returns
    -------
    file_modification : dict
        The file modification on the form

        >>> {'project_makefile_modified': str,
        ...  'project_executable_modified': str,
        ...  'project_git_sha': str,
        ...  'bout_lib_modified': str,
        ...  'bout_git_sha': str,}
    """
    file_modification = dict()
    file_modification["project_makefile_modified"] = get_modified_time(makefile_path)

    project_executable = makefile_path.parent.joinpath(exec_name)
    file_modification["project_executable_modified"] = get_modified_time(
        project_executable
    )

    file_modification["project_git_sha"] = get_git_sha(project_path)

    bout_path = get_bout_directory()
    file_modification["bout_lib_modified"] = get_modified_time(
        bout_path.joinpath("lib", "libbout++.a")
    )
    file_modification["bout_git_sha"] = get_git_sha(bout_path)

    return file_modification


def get_git_sha(path: Path) -> str:
    """
    Return the git hash.

    Parameters
    ----------
    path : Path
        Path to query the git hash

    Returns
    -------
    git_sha : str
        The git hash
    """
    try:
        result = LocalSubmitter(path).submit_command("git rev-parse HEAD")
        git_sha: str = result.stdout.decode("utf8").strip()
    # FileNotFoundError when `git` is not found
    except (FileNotFoundError, CalledProcessError) as error:
        if isinstance(error, FileNotFoundError):
            error_str = error.args[1]
        elif isinstance(error, CalledProcessError):
            error_str = error.args[2]
        else:
            error_str = "Unknown error"
        logging.warning("Could not retrieve git sha: %s", error_str)
        git_sha = "None"
    return git_sha


def get_system_info() -> Dict[str, str]:
    """
    Return the system information.

    Returns
    -------
    attributes : dict
        Dictionary with the attributes of the system
    """
    # From
    # https://stackoverflow.com/questions/11637293/iterate-over-object-attributes-in-python
    sys_info = platform.uname()
    attributes = {
        name: getattr(sys_info, name)
        for name in dir(sys_info)
        if not name.startswith("_") and not callable(getattr(sys_info, name))
    }
    return attributes



"""Contains unittests for the database writer."""


from typing import Callable

import numpy as np
from bout_runners.database.database_reader import DatabaseReader
from bout_runners.database.database_writer import DatabaseWriter


def test_db_writer(make_test_schema: Callable) -> None:
    """
    Test we can create write to the database schemas.

    Specifically this test that:
    1. We can write to the `split` table
    2. That only one record is made
    3. That the type is correct
    4. Check that the values are correct
    5. Check that it's possible to update the values

    Parameters
    ----------
    make_test_schema : function
        Function returning the database connection with the schema created
    """
    db_connection, _ = make_test_schema("write_test")
    db_reader = DatabaseReader(db_connection)

    db_writer = DatabaseWriter(db_connection)
    table_name = "split"
    dummy_split_dict = {
        "number_of_processors": 41,
        "number_of_nodes": 42,
        "processors_per_node": 43,
    }
    db_writer.create_entry(table_name, dummy_split_dict)

    # NOTE: Protected against SQL injection as table_name is hard-coded above
    table = db_reader.query(f"SELECT * FROM {table_name}")  # nosec

    # Check that the shape is expected (note that one column is
    # assigned to the id)
    assert table.shape == (1, 4)

    # Check all the elements are the same
    # https://www.quora.com/How-do-you-check-if-all-elements-in-a-NumPy-array-are-the-same-in-Python-pandas
    values = table.dtypes.values
    assert (values == np.dtype("int64")).all()

    for key, value in dummy_split_dict.items():
        assert table.loc[0, key] == value  # pylint: disable=no-member

    update_fields = ("number_of_processors", "number_of_nodes")
    search_condition = (
        f"processors_per_node = " f'{dummy_split_dict["processors_per_node"]}'
    )
    values = tuple(dummy_split_dict[field] - 10 for field in update_fields)
    db_writer.update(
        db_writer.create_update_string(update_fields, table_name, search_condition),
        values,
    )
    # NOTE: Protected against SQL injection as table_name is hard-coded above
    table = db_reader.query(f"SELECT * FROM {table_name}")  # nosec
    for index, field in enumerate(update_fields):
        # pylint: disable=no-member
        assert table.loc[:, field].values[0] == values[index]



"""Contains unittests for BOUT paths."""


from pathlib import Path
from typing import Callable

import pytest
from bout_runners.executor.bout_paths import BoutPaths


def test_bout_path(yield_conduction_path: Path, copy_bout_inp: Callable) -> None:
    """
    Test that BoutPath is copying BOUT.inp.

    Parameters
    ----------
    yield_conduction_path : Path
        Path to the BOUT++ conduction example
        See the yield_conduction_path for more details
    copy_bout_inp : function
        Function which copies BOUT.inp and returns the path to the temporary
        directory
        See the copy_bout_inp fixture for more details
    """
    project_path = yield_conduction_path
    tmp_path_name = "tmp_BoutPath_test"

    # NOTE: We use the fixture here to automatically remove the
    # directory after the test
    tmp_path_dir = copy_bout_inp(project_path, tmp_path_name)

    # We remove the BOUT.inp to verify that BoutPaths copied the file
    tmp_path_dir.joinpath("BOUT.inp").unlink()

    bout_paths = BoutPaths(project_path=project_path, bout_inp_dst_dir=tmp_path_name)

    assert project_path.joinpath(tmp_path_name, "BOUT.inp").is_file()

    with pytest.raises(FileNotFoundError):
        # NOTE: type: ignore due to https://github.com/python/mypy/issues/3004
        bout_paths.bout_inp_src_dir = "dir_without_BOUT_inp"  # type: ignore



"""Module containing the StatusChecker class."""


import logging
import time
from typing import Optional, Union
from pathlib import Path

import psutil
from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_reader import DatabaseReader
from bout_runners.log.log_reader import LogReader
from bout_runners.metadata.metadata_updater import MetadataUpdater
from pandas import DataFrame


class StatusChecker:
    r"""
    Class to check and update the status of runs.

    Attributes
    ----------
    __db_connector : DatabaseConnector
        Connection to the database under consideration
    __db_reader : DatabaseReader
        Object to read the database with
    project_path : Path
        Path to the project

    Methods
    -------
    check_and_update_status()
        Check and update the status for the schema
    check_and_update_status_until_complete()
        Check and update the status until all runs are stopped
    __check_submitted(metadata_updater, submitted_to_check)
        Check the status of all runs which has status `submitted`
    __check_running(metadata_updater, running_to_check)
        Check the status of all runs which has status `running`
     __check_if_stopped(log_reader, metadata_updater)
        Check if a run has stopped
    check_if_running_or_errored(log_reader)
        Check if a run is still running or has errored

    Examples
    --------
    >>> from pathlib import Path
    >>> from bout_runners.database.database_connector import \
    ...     DatabaseConnector
    >>> db_connector = DatabaseConnector('name_of_db')
    >>> project_path = Path('path').joinpath('to', 'project')
    >>> status_checker = StatusChecker(db_connector, project_path)
    >>> status_checker.check_and_update_status()

    Any updates to the runs will be written to the database.
    Alternatively, one can run the program until all jobs have
    stopped by calling

    >>> status_checker.check_and_update_until_complete()
    """

    def __init__(
        self,
        db_connector: Optional[DatabaseConnector] = None,
        project_path: Optional[Union[Path, str]] = None,
    ) -> None:
        """
        Set connector, reader and a project path.

        Notes
        -----
        The StatusChecker instance only checks the project belonging to the same
        database schema grouped together by the `db_connector`

        Parameters
        ----------
        db_connector : DatabaseConnector
            Connection to the database
        project_path : Path
            Path to the project (the root directory with which usually contains the
            makefile and the executable)
        """
        self.__db_connector = (
            db_connector if db_connector is not None else DatabaseConnector()
        )
        self.__db_reader = DatabaseReader(self.__db_connector)
        self.__project_path = Path(project_path) if project_path is not None else Path()

    def check_and_update_status(self) -> None:
        """
        Check and update the status for the schema.

        Raises
        ------
        RuntimeError
            If the schema does not exist
        """
        # Check that run table exist
        if not self.__db_reader.check_tables_created():
            logging.error(
                "No tables found in %s", self.__db_reader.db_connector.db_path,
            )
            message = "Can not check the status of schemas that does not exist"
            raise RuntimeError(message)

        # Create place holder metadata_updater
        metadata_updater = MetadataUpdater(self.__db_connector, run_id=-1)

        # Check runs with status 'submitted'
        query = (
            "SELECT name, id AS run_id FROM run WHERE\n"
            "latest_status = 'submitted' OR\n"
            "latest_status = 'created'"
        )
        submitted_to_check = self.__db_reader.query(query)
        self.__check_submitted(metadata_updater, submitted_to_check)

        # Check runs with status 'running'
        query = 'SELECT name, id FROM run WHERE latest_status = "running"'
        running_to_check = self.__db_reader.query(query)
        self.__check_running(metadata_updater, running_to_check)

    def check_and_update_until_complete(self, seconds_between_update: int = 5) -> None:
        """
        Check and update the status until all runs are stopped.

        Parameters
        ----------
        seconds_between_update : int
            Number of seconds before a new status check is performed
        """
        query = (
            "SELECT name, id AS run_id FROM run WHERE\n"
            "latest_status = 'submitted' OR\n"
            "latest_status = 'created' OR\n"
            "latest_status = 'running'"
        )
        while len(self.__db_reader.query(query).index) != 0:
            self.check_and_update_status()
            time.sleep(seconds_between_update)

    def __check_submitted(
        self, metadata_updater: MetadataUpdater, submitted_to_check: DataFrame
    ) -> None:
        """
        Check the status of all runs which has status `submitted`.

        Parameters
        ----------
        metadata_updater : MetadataUpdater
            Object which updates the database
        submitted_to_check : DataFrame
            DataFrame containing the `id` and `name` of the runs with status `submitted`

        Raises
        ------
        RuntimeError
            In case log_reader.started() is True and log_reader.start_time is None
        """
        for name, run_id in submitted_to_check.itertuples(index=False):
            metadata_updater.run_id = run_id

            log_path = self.__project_path.joinpath(name, "BOUT.log.0")

            if log_path.is_file():
                log_reader = LogReader(log_path)
                if log_reader.started():
                    start_time = log_reader.start_time
                    # Assert to prevent "Incompatible types in assignment" with Optional
                    if start_time is None:
                        raise RuntimeError(
                            "log_reader.start_time is None although "
                            "log_reader.started is True"
                        )
                    metadata_updater.update_start_time(start_time)
                    latest_status = self.__check_if_stopped(
                        log_reader, metadata_updater
                    )

                else:
                    # No started time is found in the log
                    latest_status = self.check_if_running_or_errored(log_reader)
            else:
                # No log file exists
                # NOTE: This means that the execution is either in a
                #       queue or has failed the submission.
                #       For now, we still consider this as submitted
                latest_status = "submitted"

            metadata_updater.update_latest_status(latest_status)

    def __check_running(
        self, metadata_updater: MetadataUpdater, running_to_check: DataFrame
    ) -> None:
        """
        Check the status of all runs which has status `running`.

        Parameters
        ----------
        metadata_updater : MetadataUpdater
            Object which updates the database
        running_to_check : DataFrame
            DataFrame containing the `id` and `name` of the runs with status `running`
        """
        for name, run_id in running_to_check.itertuples(index=False):
            metadata_updater.run_id = run_id
            log_path = self.__project_path.joinpath(name, "BOUT.log.0")
            log_reader = LogReader(log_path)
            latest_status = self.check_if_running_or_errored(log_reader)
            metadata_updater.update_latest_status(latest_status)

    def __check_if_stopped(
        self, log_reader: LogReader, metadata_updater: MetadataUpdater
    ) -> str:
        """
        Check if a run has stopped.

        Parameters
        ----------
        log_reader : LogReader
            The object which reads log files
        metadata_updater : MetadataUpdater
            Object which updates the database

        Returns
        -------
        latest_status : str
            The latest status

        Raises
        ------
        RuntimeError
            In case log_reader.ended() is True and log_reader.end_time is None
        """
        if log_reader.ended():
            end_time = log_reader.end_time
            # Assert to prevent "Incompatible types in assignment" with Optional
            if end_time is None:
                raise RuntimeError(
                    "log_reader.end_time is None although " "log_reader.ended() is True"
                )
            metadata_updater.update_stop_time(end_time)
            latest_status = "complete"
        else:
            latest_status = self.check_if_running_or_errored(log_reader)
        return latest_status

    @staticmethod
    def check_if_running_or_errored(log_reader: LogReader) -> str:
        """
        Check if a run is still running or has errored.

        Parameters
        ----------
        log_reader : LogReader
            The object which reads log files

        Returns
        -------
        latest_status : str
            The latest status
        """
        pid = log_reader.pid
        if pid is None:
            latest_status = "created"
        elif psutil.pid_exists(pid):
            latest_status = "running"
        else:
            latest_status = "error"
        return latest_status



"""Contains unittests for the database creator."""


import sqlite3
from typing import Callable

import pytest
from bout_runners.database.database_creator import DatabaseCreator
from bout_runners.database.database_reader import DatabaseReader


def test_db_creator(make_test_database: Callable, make_test_schema: Callable) -> None:
    """
    Test we can create the database schemas.

    Specifically this test that:
    1. The database is empty on creation
    2. The tables are created
    3. It is not possible to create the schema more than once
    4. Check that all expected tables have been created

    Parameters
    ----------
    make_test_database : function
        Function returning the database connection
    make_test_schema : function
        Function returning the database connection and the final parameters as sql types
    """
    db_connection_no_schema = make_test_database("test_creation_without_schema")
    db_reader_no_schema = DatabaseReader(db_connection_no_schema)

    # There should be no tables before creating them
    assert not db_reader_no_schema.check_tables_created()

    db_connection_schema, final_parameters_as_sql_types = make_test_schema(
        "test_creation_with_schema"
    )
    db_reader_schema = DatabaseReader(db_connection_schema)
    db_creator = DatabaseCreator(db_connection_schema)

    # The tables should now have been created
    assert db_reader_schema.check_tables_created()

    with pytest.raises(sqlite3.OperationalError):
        db_creator.create_all_schema_tables(final_parameters_as_sql_types)

    # Check that all tables has been created
    non_parameter_tables = {
        "system_info",
        "split",
        "file_modification",
        "parameters",
        "run",
    }
    parameter_tables = set(
        el.replace(":", "_") for el in final_parameters_as_sql_types.keys()
    )
    query_str = 'SELECT name FROM sqlite_master WHERE type="table"'
    table = db_reader_schema.query(query_str)

    actual = table.loc[:, "name"].values  # pylint: disable=no-member
    assert non_parameter_tables.union(parameter_tables) == set(actual)


def test_get_create_table_statement() -> None:
    """Test that get_create_table_statement returns expected output."""
    result = DatabaseCreator.get_create_table_statement(
        table_name="foo",
        columns=dict(bar="baz", foobar="qux"),
        primary_key="quux",
        foreign_keys=dict(quuz=("corge", "grault"), garply=("waldo", "fred")),
    )

    expected = (
        "CREATE TABLE foo \n"
        "(   quux INTEGER PRIMARY KEY,\n"
        "    bar baz NOT NULL,\n"
        "    foobar qux NOT NULL,\n"
        "    quuz INTEGER NOT NULL,\n"
        "    garply INTEGER NOT NULL,\n"
        "    FOREIGN KEY(quuz) \n"
        "        REFERENCES corge(grault)\n"
        "            ON UPDATE CASCADE\n"
        "            ON DELETE CASCADE,\n"
        "    FOREIGN KEY(garply) \n"
        "        REFERENCES waldo(fred)\n"
        "            ON UPDATE CASCADE\n"
        "            ON DELETE CASCADE)"
    )

    assert result == expected



"""Contains unittests for the config module."""


from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch
from bout_runners.configure_bout_runners import (
    set_bout_directory,
    set_log_file_directory,
    set_log_level,
)
from bout_runners.utils.logs import get_log_config
from bout_runners.utils.paths import (
    get_bout_directory,
    get_bout_runners_configuration,
    get_log_file_directory,
)


def test_set_log_level(get_mock_config_path: Path, monkeypatch: MonkeyPatch) -> None:
    """
    Test that the log level is changeable.

    Parameters
    ----------
    get_mock_config_path : Path
        The mocked config directory
    monkeypatch : MonkeyPatch
        MonkeyPatch from pytest
    """
    # Test with parameter input
    _ = get_mock_config_path
    level = "CRITICAL"
    level_number = 5
    set_log_level(level)
    config = get_log_config()
    assert config["handlers"]["file_handler"]["level"] == level
    assert config["handlers"]["console_handler"]["level"] == level
    assert config["root"]["level"] == level

    # Test with incorrect input
    with pytest.raises(ValueError):
        set_log_level("Not a level")

    # Test with empty input
    monkeypatch.setattr("builtins.input", lambda _: "")
    set_log_level()
    config = get_log_config()
    assert config["handlers"]["file_handler"]["level"] == level
    assert config["handlers"]["console_handler"]["level"] == level
    assert config["root"]["level"] == level

    # Test with non-empty input
    level = "ERROR"
    level_number = 3
    monkeypatch.setattr("builtins.input", lambda _: level_number)
    set_log_level()
    config = get_log_config()
    assert config["handlers"]["file_handler"]["level"] == level
    assert config["handlers"]["console_handler"]["level"] == level
    assert config["root"]["level"] == level


def test_set_log_file_directory(
    get_mock_config_path: Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Test that the log file directory is changeable.

    Parameters
    ----------
    get_mock_config_path : Path
        The mocked config directory
    monkeypatch : MonkeyPatch
        MonkeyPatch from pytest
    """
    config_path = get_mock_config_path
    original_dir = get_log_file_directory()

    # Test with empty input
    monkeypatch.setattr("builtins.input", lambda _: "")
    set_log_file_directory()
    config = get_bout_runners_configuration()
    assert config["log"]["directory"] == str(original_dir)

    # Test with parameter input
    log_dir = config_path.joinpath("test_with_parameter")
    set_log_file_directory(log_dir)
    config = get_bout_runners_configuration()
    assert config["log"]["directory"] == str(log_dir)

    # Test with non-empty input
    log_dir = config_path.joinpath("test_with_parameter_2")
    monkeypatch.setattr("builtins.input", lambda _: str(log_dir))
    set_log_file_directory()
    config = get_bout_runners_configuration()
    assert config["log"]["directory"] == str(log_dir)


def test_set_bout_directory(
    get_mock_config_path: Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Test that the BOUT++ directory is changeable.

    Parameters
    ----------
    get_mock_config_path : Path
        The mocked config directory
    monkeypatch : MonkeyPatch
        MonkeyPatch from pytest
    """
    _ = get_mock_config_path
    original_dir = get_bout_directory()

    # Test with empty input
    monkeypatch.setattr("builtins.input", lambda _: "")
    set_bout_directory()
    config = get_bout_runners_configuration()
    assert config["bout++"]["directory"] == str(original_dir)

    # Test with incorrect input
    with pytest.raises(ValueError):
        set_bout_directory(original_dir.joinpath("not", "a", "dir"))

    # Test with parameter input
    bout_dir = original_dir.parent
    set_bout_directory(bout_dir)
    config = get_bout_runners_configuration()
    assert config["bout++"]["directory"] == str(bout_dir)

    # Test with non-empty input
    monkeypatch.setattr("builtins.input", lambda _: str(original_dir))
    set_bout_directory()
    config = get_bout_runners_configuration()
    assert config["bout++"]["directory"] == str(original_dir)



"""Contains unittests for default parameters."""


from pathlib import Path
from typing import Callable

from bout_runners.parameters.default_parameters import DefaultParameters


def test_default_parameters(
    get_test_data_path: Path, yield_bout_path_conduction: Callable
) -> None:
    """
    Test the DefaultParameter.

    Test by:
    1. Checking that setting only the `settings_path` parameter yields a dict
    2. Checking that setting only the `bout_paths` parameter yields a dict

    Parameters
    ----------
    get_test_data_path : Path
        Path to the test data
    yield_bout_path_conduction : function
        Function returning the bout_paths object
    """
    settings_path = get_test_data_path.joinpath("BOUT.settings")
    default_parameters_settings = DefaultParameters(settings_path=settings_path)
    settings_dict = default_parameters_settings.get_default_parameters()

    assert isinstance(settings_dict, dict)

    bout_paths = yield_bout_path_conduction("test_default_parameters")
    default_parameters_bout_paths = DefaultParameters(bout_paths=bout_paths)
    bout_paths_dict = default_parameters_bout_paths.get_default_parameters()

    assert isinstance(bout_paths_dict, dict)



"""Contains unittests for the database connector."""


import sqlite3
from typing import Callable

import pytest


def test_db_connector(make_test_database: Callable) -> None:
    """
    Test the connection.

    This will be done by:
    1. Check that the connection is open
    2. Check that one can execute a statement
    3. Check that it raises error on an invalid statement
    4. Check that it is not possible to change the db_path
    4. Check that it is not possible to change the connection

    Parameters
    ----------
    make_test_database : function
        Function which returns the database connection
    """
    db_connection = make_test_database("connection_test")
    assert isinstance(db_connection.connection, sqlite3.Connection)

    db_connection.execute_statement("CREATE TABLE my_table (col INT)")
    db_connection.execute_statement("SELECT 1+1")

    with pytest.raises(sqlite3.OperationalError):
        db_connection.execute_statement("THIS IS AN INVALID STATEMENT")

    with pytest.raises(AttributeError):
        db_connection.db_path = "invalid"

    with pytest.raises(AttributeError):
        db_connection.connection = "invalid"



"""Contains the class dealing with the default parameters."""


import ast
import configparser
import logging
import re
from pathlib import Path
from typing import Dict, Optional, Union

from bout_runners.executor.bout_paths import BoutPaths
from bout_runners.executor.executor import Executor
from bout_runners.parameters.run_parameters import RunParameters
from bout_runners.submitter.local_submitter import LocalSubmitter


class DefaultParameters:
    """
    Class which deals with the default parameters.

    The default parameters are those set internally in BOUT++, or in the specified
    BOUT.inp file

    Attributes
    ----------
    self.__bout_paths : BoutPaths
        Object for the BOUT++ paths
    self.__settings_path : None or Path
        Path to the BOUT.settings file

    Methods
    -------
    run_parameters_run()
        Execute a run to obtain the default parameters.
    get_default_parameters()
        Return the default parameters from the settings file.

    Examples
    --------
    The easiest way to use DefaultParameters is to run a script from the root
    directory of the project (i.e. where the `Makefile` and `data` directory are
    normally situated. The script can simply call

    >>> DefaultParameters().get_default_parameters()
    {'global': {'append': False, 'async_send': False, ...}}

    A more elaborate example where all the dependency objects are
    built manually:

    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Get the default parameters

    >>> default_parameter = DefaultParameters(bout_paths=bout_paths)
    >>> default_parameter.get_default_parameters()
    {'global': {'append': False, 'async_send': False, ...}}
    """

    def __init__(
        self,
        bout_paths: Optional[BoutPaths] = None,
        settings_path: Optional[Path] = None,
    ) -> None:
        """
        Set the member data.

        If the settings_path is None, the constructor will call run_parameters_run to
        create a settings_path

        Warnings
        --------
        There can be a potential mismatch between a user provided `settings_path` and
        the actual default values. This can occur if the user has updated `BOUT.inp`
        without updating the `BOUT.settings` file.
        It is therefore recommended to set `settings_path` to None unless the user is
        sure the `BOUT.settings` file pointed to by `settings_path` is up to date

        Parameters
        ----------
        bout_paths : BoutPaths or None
            Object containing the paths of the project
            Will only be used in the `run_parameters_run` call if the `settings_path`
            is not valid
        settings_path : None or Path
            Path to the up-to-date `settings_path`
            Will invoke `run_parameters_run` if set to None
        """
        self.__bout_paths = bout_paths
        self.__settings_path = Path() if settings_path is None else Path(settings_path)

        if not self.__settings_path.is_file():
            logging.info(
                "Running parameter run as the parameters of " "the project are unknown"
            )
            self.run_parameters_run(self.__bout_paths)

    def run_parameters_run(self, bout_paths: Optional[BoutPaths]) -> None:
        """
        Execute a run to obtain the default parameters.

        A settings run executes the executable of the project with nout = 0 in order
        to capture all parameters used in the project

        Parameters
        ----------
        bout_paths : BoutPaths
            Object containing the paths of the project
        """
        if bout_paths is None:
            bout_paths = BoutPaths(bout_inp_dst_dir="settings_run")
        else:
            # NOTE: type: ignore due to https://github.com/python/mypy/issues/3004
            bout_paths.bout_inp_dst_dir = "settings_run"  # type: ignore

        executor = self.get_test_executor(bout_paths)

        executor.execute()

        self.__settings_path = bout_paths.bout_inp_dst_dir.joinpath("BOUT.settings")

    @staticmethod
    def get_test_executor(bout_paths: BoutPaths) -> Executor:
        """
        Return the executor used for test (i.e. where nout=0).

        Parameters
        ----------
        bout_paths : BoutPaths
            Object containing the BOUT++ paths

        Returns
        -------
        executor : Executor
            Executor instantiated with the test set up
        """
        run_parameters = RunParameters({"global": {"nout": 0}})
        executor = Executor(
            bout_paths=bout_paths,
            submitter=LocalSubmitter(bout_paths.project_path),
            run_parameters=run_parameters,
        )
        return executor

    def get_default_parameters(
        self,
    ) -> Dict[str, Dict[str, Union[str, int, float, bool]]]:
        """
        Return the default parameters from the settings file.

        Returns
        -------
        default_parameters_dict : dict
            Dictionary containing the parameters given in BOUT.settings
            On the form

            >>> {'section': {'parameter': 'value'}}

        Notes
        -----
        1. The section-less part of BOUT.settings will be renamed `global`
        2. In the `global` section, the keys `d` and the directory to the BOUT.inp
           file will be removed
        3. If the section `all` is present in BOUT.settings, the section will be
           renamed `all_boundaries` as `all` is a protected SQL keyword
        4. The section `run` will be dropped due to bout_runners own `run` table
        5. The string values will be stored using lowercase
        """
        # The settings file lacks a header for the global parameter
        # Therefore, we add add the header [global]
        with self.__settings_path.open("r") as settings_file:
            settings_memory = f"[global]\n{settings_file.read()}"

        config = configparser.ConfigParser()
        config.read_string(settings_memory)

        default_parameters_dict: Dict[str, Dict[str, Union[str, int, float]]] = dict()

        for section in config.sections():
            default_parameters_dict[section] = dict()
            for key, val in config[section].items():
                # Strip comments
                capture_all_but_comment = "^([^#]*)"
                matches = re.findall(capture_all_but_comment, val, re.M)

                # Exclude comment line
                if len(matches) == 0:
                    continue

                # Capitalize in case of boolean
                stripped_val = matches[0].capitalize()

                # If type is not found, type is str
                try:
                    val = ast.literal_eval(stripped_val)
                except (SyntaxError, ValueError):
                    # Strip the whitespaces and cast to lowercase
                    val = stripped_val.strip().lower()

                default_parameters_dict[section][key] = val

        # NOTE: Bug in .settings: -d path is captured with # not in use
        bout_inp_dir = self.__settings_path.parent
        default_parameters_dict["global"].pop("d", None)
        default_parameters_dict["global"].pop(str(bout_inp_dir).lower(), None)

        if "all" in default_parameters_dict.keys():
            default_parameters_dict["all_boundaries"] = default_parameters_dict.pop(
                "all"
            )

        # Drop run as bout_runners will make its own table with that
        # name
        default_parameters_dict.pop("run", None)

        return default_parameters_dict



"""Contains unittests for final parameters."""

from bout_runners.parameters.default_parameters import DefaultParameters
from bout_runners.parameters.final_parameters import FinalParameters
from bout_runners.parameters.run_parameters import RunParameters


def test_final_parameters(get_default_parameters: DefaultParameters) -> None:
    """
    Test that RunParameters overwrites DefaultParameters.

    Parameters
    ----------
    get_default_parameters : DefaultParameters
        The DefaultParameters object
    """
    default_parameters = get_default_parameters
    run_parameters = RunParameters({"global": {"timestep": False}})
    final_parameters = FinalParameters(default_parameters, run_parameters)

    final_parameters_dict = final_parameters.get_final_parameters()

    assert final_parameters_dict["global"]["timestep"] == 0


def test_cast_parameters_to_sql_type(get_default_parameters: DefaultParameters) -> None:
    """
    Test that obtain_project_parameters returns expected output.

    Parameters
    ----------
    get_default_parameters : DefaultParameters
        The DefaultParameters object
    """
    default_parameters = get_default_parameters
    final_parameters = FinalParameters(default_parameters)
    final_parameters_dict = final_parameters.get_final_parameters()
    parameter_as_sql = final_parameters.cast_to_sql_type(final_parameters_dict)
    assert isinstance(parameter_as_sql, dict)
    assert "global" in parameter_as_sql.keys()
    assert isinstance(parameter_as_sql["global"], dict)
    assert parameter_as_sql["global"]["append"] == "INTEGER"  # bool
    assert parameter_as_sql["conduction"]["chi"] == "REAL"  # float
    assert parameter_as_sql["global"]["mxg"] == "INTEGER"  # int
    assert parameter_as_sql["global"]["datadir"] == "TEXT"  # str



"""Contains unittests for the StatusChecker."""

from datetime import datetime
from pathlib import Path
from typing import Callable

import pytest
from bout_runners.database.database_reader import DatabaseReader
from bout_runners.metadata.status_checker import StatusChecker


def test_status_checker_run_time_error(make_test_database: Callable) -> None:
    """
    Test that the status checker raises RuntimeError without tables.

    Parameters
    ----------
    make_test_database : DatabaseConnector
        Connection to the test database
    """
    db_connector = make_test_database("status_checker_no_table")
    status_checker = StatusChecker(db_connector, Path())

    with pytest.raises(RuntimeError):
        status_checker.check_and_update_status()


@pytest.mark.parametrize(
    "test_case",
    (
        "no_log_file_no_pid_not_started_not_ended_no_mock_pid_submitted",
        "log_file_no_pid_not_started_not_ended_no_mock_pid_created",
        "log_file_pid_not_started_not_ended_no_mock_pid_error",
        "log_file_pid_not_started_not_ended_mock_pid_running",
        "log_file_pid_started_not_ended_no_mock_pid_error",
        "log_file_pid_started_not_ended_mock_pid_running",
        "log_file_pid_started_ended_no_mock_pid_error",
        "log_file_pid_started_ended_no_mock_pid_complete",
    ),
)
def test_status_checker(
    test_case: str,
    get_test_data_path: Path,
    get_test_db_copy: Callable,
    mock_pid_exists: Callable,
    copy_test_case_log_file: Callable,
) -> None:
    """
    Test the StatusChecker exhaustively (excluding raises and loop).

    Parameters
    ----------
    test_case : str
        Description of the test on the form

        >>> ('<log_file_present>_<pid_present_in_log>_'
        ...  '<started_time_present_in_log>_<ended_time_present_in_log>'
        ...  '_<whether_pid_exists>_<new_status>')

    get_test_data_path : Path
        Path to test data
    get_test_db_copy : function
        Function which returns a a database connector to the copy of the
        test database
    mock_pid_exists : function
        Function which sets up a monkeypatch for psutil.pid_exist
    copy_test_case_log_file : function
        Function which copies log files according to the test_case
    """
    project_path = get_test_data_path
    db_connector = get_test_db_copy(test_case)
    mock_pid_exists(test_case)
    copy_test_case_log_file(test_case)

    db_reader = DatabaseReader(db_connector)

    status_checker = StatusChecker(db_connector, project_path)
    status_checker.check_and_update_status()

    # Check that the correct status has been assigned to "running"
    # pylint: disable=no-member
    result = db_reader.query(
        "SELECT latest_status FROM run WHERE name = " "'testdata_5'"
    ).loc[0, "latest_status"]
    assert result == "running"

    # Check that the correct status has been assigned to "submitted"
    expected = test_case.split("_")[-1]
    # pylint: disable=no-member
    result = db_reader.query(
        "SELECT latest_status FROM run WHERE name = " "'testdata_6'"
    ).loc[0, "latest_status"]
    assert result == expected

    # Check that correct start_time has been set
    if "not_started" not in test_case:
        expected = str(datetime(2020, 5, 1, 17, 7, 10))
        # pylint: disable=no-member
        result = db_reader.query(
            "SELECT start_time FROM run WHERE name = " "'testdata_6'"
        ).loc[0, "start_time"]
        assert expected == result

    # Check that correct end_time has been set
    if "not_ended" not in test_case and "complete" in test_case:
        expected = str(datetime(2020, 5, 1, 17, 7, 14))
        # pylint: disable=no-member
        result = db_reader.query(
            "SELECT stop_time FROM run WHERE name = " "'testdata_6'"
        ).loc[0, "stop_time"]
        assert expected == result


@pytest.mark.timeout(60)
def test_status_checker_until_complete_infinite(
    get_test_data_path: Path,
    get_test_db_copy: Callable,
    copy_test_case_log_file: Callable,
) -> None:
    """
    Test the infinite loop of StatusChecker.

    Parameters
    ----------
    get_test_data_path : Path
        Path to the test data
    get_test_db_copy : function
        Function which returns a DatabaseConnector connected to a copy of test.db
    copy_test_case_log_file : function
        Return the function for copying the test case log files
    """
    test_case = "infinite_log_file_pid_started_ended_no_mock_pid_complete"

    project_path = get_test_data_path
    db_connector = get_test_db_copy(test_case)
    copy_test_case_log_file(test_case)

    # Remove row which has status running (as it will always have
    # this status)
    db_connector.execute_statement("DELETE FROM run WHERE name = 'testdata_5'")

    db_reader = DatabaseReader(db_connector)

    status_checker = StatusChecker(db_connector, project_path)
    status_checker.check_and_update_until_complete()

    query = (
        "SELECT name, id AS run_id FROM run WHERE\n"
        "latest_status = 'submitted' OR\n"
        "latest_status = 'created' OR\n"
        "latest_status = 'running'"
    )
    assert len(db_reader.query(query).index) == 0



"""Contains unittests for the log_reader."""


from datetime import datetime
from pathlib import Path
from typing import Dict

from bout_runners.log.log_reader import LogReader


def test_started(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.started.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.started()

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.started()

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.started() is False


def test_ended(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.ended.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.ended()

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.ended() is False

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.ended() is False


def test_pid_exist(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.pid_exist.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.pid_exist()

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.pid_exist()

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.ended() is False


def test_start_time(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.start_time.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    start_time = datetime(2020, 5, 1, 17, 7, 10)
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.start_time == start_time

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.start_time == start_time

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.start_time is None


def test_end_time(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.end_time.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    end_time = datetime(2020, 5, 1, 17, 7, 14)
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.end_time == end_time

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.end_time is None

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.end_time is None


def test_pid(yield_logs: Dict[str, Path]) -> None:
    """
    Test self.pid.

    Parameters
    ----------
    yield_logs : dict of Path
        A dictionary containing the log paths used for testing
    """
    log_paths = yield_logs
    success_log_reader = LogReader(log_paths["success_log"])
    assert success_log_reader.pid == 1191

    failed_log_reader = LogReader(log_paths["fail_log"])
    assert failed_log_reader.pid == 1190

    unfinished_no_pid_log_reader = LogReader(log_paths["unfinished_no_pid_log"])
    assert unfinished_no_pid_log_reader.pid is None



"""Module containing the MetadataRecorder class."""


from datetime import datetime
from typing import Dict, Mapping, Optional, Union

from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_reader import DatabaseReader
from bout_runners.database.database_utils import get_file_modification, get_system_info
from bout_runners.database.database_writer import DatabaseWriter
from bout_runners.executor.bout_paths import BoutPaths
from bout_runners.make.make import Make
from bout_runners.parameters.final_parameters import FinalParameters
from bout_runners.submitter.processor_split import ProcessorSplit


class MetadataRecorder:
    r"""
    Class for recording the metadata of the runs.

    Attributes
    ----------
    __db_writer : DatabaseWriter
        Getter variable for db_writer
    __db_reader : DatabaseReader
        Getter variable for db_reader
    db_writer : DatabaseWriter
        Object which writes to the database
    db_reader : DatabaseReader
        Object which reads from the database

    Methods
    -------
    capture_new_data_from_run(runner, processor_split)
        Capture new data from a run
    _create_parameter_tables_entry(parameters_dict)
        Insert the parameters into a the parameter tables

    Examples
    --------
    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.parameters.default_parameters import DefaultParameters
    >>> from bout_runners.parameters.final_parameters import FinalParameters
    >>> from bout_runners.database.database_connector import DatabaseConnector
    >>> from bout_runners.submitter.processor_split import ProcessorSplit

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination','BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Obtain the parameters

    >>> default_parameters = DefaultParameters(bout_paths)
    >>> final_parameters = FinalParameters(default_parameters)
    >>> final_parameters_dict = final_parameters.get_final_parameters()
    >>> final_parameters_as_sql_types = \
    ...     final_parameters.cast_to_sql_type(
    ...     final_parameters_dict)

    Create the metadata recorder object

    >>> db_connection = DatabaseConnector('name')
    >>> metadata_recorder = MetadataRecorder(db_connection,
    ...                                      bout_paths,
    ...                                      final_parameters)

    Capture the data to the database

    >>> metadata_recorder.capture_new_data_from_run(ProcessorSplit())
    None
    """

    def __init__(
        self,
        db_connector: DatabaseConnector,
        bout_paths: BoutPaths,
        final_parameters: FinalParameters,
    ) -> None:
        """
        Set the database to use.

        Parameters
        ----------
        db_connector : DatabaseConnector
            The database connector
        bout_paths : BoutPaths
            Object containing the paths
        final_parameters : FinalParameters
            Object containing the final parameters
        """
        self.__db_writer = DatabaseWriter(db_connector)
        self.__db_reader = DatabaseReader(db_connector)
        self.__bout_paths = bout_paths
        self.__final_parameters = final_parameters
        self.__make = Make(self.__bout_paths.project_path)

    @property
    def db_reader(self) -> DatabaseReader:
        """
        Set the properties of self.db_reader.

        Returns
        -------
        self.__db_reader : DatabaseReader
            The database reader object

        Notes
        -----
        The db_reader is read only
        """
        return self.__db_reader

    @property
    def db_writer(self):
        """
        Set the properties of self.db_writer.

        Returns
        -------
        self.__db_writer : DatabaseWriter
            The database writer object

        Notes
        -----
        The db_writer is read only
        """
        return self.__db_writer

    def capture_new_data_from_run(
        self, processor_split: ProcessorSplit, force: bool = False
    ) -> Optional[int]:
        """
        Capture new data from a run.

        This function will capture all uncaptured data from a run.
        If all data has been captured previously, it means that the run has already
        been executed, and new_entry = False will be returned.

        Parameters
        ----------
        processor_split : ProcessorSplit
            The processor split object
        force : bool
            Store entry to the run table even if a entry with the same parameter exists
            This will typically be used if the bout_runners is forcefully executing
            a run

        Returns
        -------
        run_id : None or int
            If no previous run with the same configuration has been executed,
            this will return None, else the run_id is returned
        """
        # Initiate the run_dict (will be filled with the ids)
        run_dict: Dict[str, Union[str, int, float, None]] = {
            "name": self.__bout_paths.bout_inp_dst_dir.name
        }

        # Update the parameters
        parameters_dict = self.__final_parameters.get_final_parameters()

        run_dict["parameters_id"] = self._create_parameter_tables_entry(parameters_dict)

        # Update the file_modification
        file_modification_dict = get_file_modification(
            self.__bout_paths.project_path,
            self.__make.makefile_path,
            self.__make.exec_name,
        )
        run_dict["file_modification_id"] = self.__db_reader.get_entry_id(
            "file_modification", file_modification_dict
        )
        if run_dict["file_modification_id"] is None:
            run_dict["file_modification_id"] = self.create_entry(
                "file_modification", file_modification_dict
            )

        # Update the split
        split_dict = {
            "number_of_processors": processor_split.number_of_processors,
            "number_of_nodes": processor_split.number_of_nodes,
            "processors_per_node": processor_split.processors_per_node,
        }
        run_dict["split_id"] = self.__db_reader.get_entry_id("split", split_dict)
        if run_dict["split_id"] is None:
            run_dict["split_id"] = self.create_entry("split", split_dict)

        # Update the system info
        system_info_dict = get_system_info()
        run_dict["system_info_id"] = self.__db_reader.get_entry_id(
            "system_info", system_info_dict
        )
        if run_dict["system_info_id"] is None:
            run_dict["system_info_id"] = self.create_entry(
                "system_info", system_info_dict
            )

        # Update the run
        run_id = self.__db_reader.get_entry_id("run", run_dict)
        if force or run_id is None:
            run_dict["latest_status"] = "submitted"
            run_dict["submitted_time"] = datetime.now().isoformat()
            _ = self.create_entry("run", run_dict)

        return run_id

    def create_entry(
        self, table_name: str, entries_dict: Mapping[str, Union[int, str, float, None]]
    ) -> int:
        """
        Create a database entry and return the entry id.

        Parameters
        ----------
        table_name : str
            Name of the table
        entries_dict : dict
            Dictionary containing the entries as key value pairs

        Returns
        -------
        entry_id : int
            The id of the newly created entry

        Raises
        ------
        RuntimeError
            If the newly created id could not be fetched
        """
        self.__db_writer.create_entry(table_name, entries_dict)
        entry_id = self.__db_reader.get_entry_id(table_name, entries_dict)
        if entry_id is None:
            raise RuntimeError("Could not fetch the newly created id")
        return entry_id

    def _create_parameter_tables_entry(
        self, parameters_dict: Dict[str, Dict[str, Union[int, str, float]]]
    ) -> int:
        """
        Insert the parameters into a the parameter tables.

        Parameters
        ----------
        parameters_dict : dict
            The dictionary on the form

            >>> {'section': {'parameter': 'value'}}

        Returns
        -------
        parameters_id : int
            The id key from the `parameters` table

        Notes
        -----
        All `:` will be replaced by `_` in the section names
        """
        parameters_foreign_keys = dict()
        parameter_sections = list(parameters_dict.keys())

        for section in parameter_sections:
            # Replace bad characters for SQL
            section_name = section.replace(":", "_")
            section_parameters = parameters_dict[section]
            section_id = self.__db_reader.get_entry_id(section_name, section_parameters)
            if section_id is None:
                section_id = self.create_entry(section_name, section_parameters)

            parameters_foreign_keys[f"{section_name}_id"] = section_id

        # Update the parameters table
        parameters_id = self.__db_reader.get_entry_id(
            "parameters", parameters_foreign_keys
        )
        if parameters_id is None:
            parameters_id = self.create_entry("parameters", parameters_foreign_keys)

        return parameters_id



"""Contains unittests for the names module."""


from pathlib import Path
from typing import Optional

import pytest
from bout_runners.utils.names import get_exec_name, get_makefile_name, get_makefile_path


@pytest.mark.parametrize(
    "filename,expected",
    [(None, "executable_bout_model"), ("Makefile_without_target", "bout_model")],
)
def test_get_exec_name(
    filename: Optional[str], expected: str, get_test_data_path: Path
) -> None:
    """
    Test that the exec name is retrievable from the makefiles.

    Parameters
    ----------
    filename : str
        Name of the Makefile
    expected : str
        Expected name of the executable
    get_test_data_path : Path
        Path to the test data
    """
    makefile_path = get_makefile_path(get_test_data_path, makefile_name=filename)
    exec_name = get_exec_name(makefile_path)
    assert exec_name == expected


def test_get_makefile_name(get_test_data_path: Path) -> None:
    """
    Test that it is possible to find a makefile name.

    Parameters
    ----------
    get_test_data_path : Path
        Path to the test data
    """
    makefile_name = get_makefile_name(get_test_data_path)
    assert makefile_name == "Makefile"


def test_get_makefile_raises(copy_makefile: Path) -> None:
    """
    Test that get_makefile_name properly raises FileNotFoundError.

    Parameters
    ----------
    copy_makefile : Path
        Path to the temporary Makefile
    """
    tmp_path = copy_makefile

    with pytest.raises(FileNotFoundError):
        get_makefile_name(tmp_path)



"""Module containing the DatabaseBase class."""


import logging
import sqlite3
from pathlib import Path
from sqlite3 import Connection
from typing import Optional

from bout_runners.utils.file_operations import get_caller_dir


class DatabaseConnector:
    """
    Class creating the database path and executing sql statements.

    Attributes
    ----------
    __db_path : None or Path
        Getter variable for db_path
    __connection : Connection
        Getter variable for connection
    db_path : Path
        Path to database
    connection : Connection
        The connection to the database

    Methods
    -------
    create_db_path(name, db_root_path)
        Create the database path
    execute_statement(sql_statement, *parameters)
        Execute a statement in the database

    Examples
    --------
    >>> database = DatabaseConnector('test')
    >>> database.execute_statement('CREATE TABLE my_table (col INT)')
    """

    def __init__(
        self, name: Optional[str] = None, db_root_path: Optional[Path] = None
    ) -> None:
        """
        Set the path to the data base.

        Parameters
        ----------
        name : str or None
            Name of the database (excluding .db)
            If set to None, the name of the caller directory will be used
        db_root_path : Path or str or None
            Path to database
            If None is set, the path will be set to $HOME/BOUT_db
        """
        # Set the database path
        self.__db_path = self.create_db_path(name, db_root_path)
        logging.info("db_path set to %s", self.db_path)

        # Open the connection
        self.__connection = sqlite3.connect(str(self.db_path))

    def __del__(self) -> None:
        """Close the connection."""
        self.__connection.close()

    @property
    def db_path(self) -> Path:
        """
        Get the properties of self.db_path.

        Returns
        -------
        self.__db_path : Path
            Absolute path to the database

        Notes
        -----
        To avoid corrupting data between databases, the setting this parameter
        outside the constructor is disabled
        """
        return self.__db_path

    @property
    def connection(self) -> Connection:
        """
        Get the properties of self.connection.

        Returns
        -------
        self.__connection : Connection
            The connection to the database

        Notes
        -----
        To avoid corrupting data between databases, the setting this parameter
        outside the constructor is disabled
        """
        return self.__connection

    @staticmethod
    def create_db_path(name: Optional[str], db_root_path: Optional[Path]) -> Path:
        """
        Create the database path.

        Parameters
        ----------
        name : str or None
            Name of the database (excluding .db)
            If set to None, the name of the caller directory will be used
        db_root_path : Path or str or None
            Path to database
            If None is set, the path will be set to $HOME/BOUT_db

        Returns
        -------
        db_path : Path
            Path to the database
        """
        if name is None:
            name = get_caller_dir().name

        if db_root_path is None:
            db_root_path = Path().home().joinpath("BOUT_db")

        db_root_path = Path(db_root_path)

        db_root_path.mkdir(exist_ok=True, parents=True)
        # NOTE: sqlite does not support schemas (except through an
        #       ephemeral ATTACH connection)
        #       Thus we will make one database per project
        # https://www.sqlite.org/lang_attach.html
        # https://stackoverflow.com/questions/30897377/python-sqlite3-create-a-schema-without-having-to-use-a-second-database
        db_path = db_root_path.joinpath(f"{name}.db")

        return db_path

    def execute_statement(self, sql_statement: str, *parameters) -> None:
        """
        Execute a statement in the database.

        Parameters
        ----------
        sql_statement : str
            The statement execute
        parameters : tuple
            Parameters used in .execute of the cursor (like )
        """
        cursor = self.__connection.cursor()
        cursor.execute(sql_statement, parameters)
        self.__connection.commit()



# pylint: skip-file
# NOTE: Ideally we'd ignore this file in pylintrc, but there seem to be an open issue
#       https://github.com/PyCQA/pylint/issues/2686
"""
Configuration file for the Sphinx documentation builder.

References
----------
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import re
import sys
from pathlib import Path

root_path = Path(__file__).absolute().parents[2]
sys.path.insert(0, str(root_path))


# Project information
project = "bout_runners"
author = "Løiten, Michael"
copyright = f"2019; {author}"
# Automatic detection of release
init_path = root_path.joinpath(project, "__init__.py")
with init_path.open("r") as f:
    version_file = f.read()
    release_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if release_match:
        release = release_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


# General configuration
master_doc = "index"  # Where sphinx should look for the documentation
extensions = [
    "sphinx.ext.autosummary",  # Summary of documentation
    "sphinx.ext.viewcode",  # Add [source] link
    "numpydoc",  # Convert numpy to rst
    "sphinx_rtd_theme",  # Gives read the docs theme
]
exclude_patterns = []  # When looking for source files
templates_path = ["_templates"]  # Templates relative to this direction
pygments_style = "friendly"  # How code is rendered

# Extensions
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_default_options
autodoc_default_options = {
    "members": True,  # Include auto documentation of members
    "inherited-members": True,  # Include inherited members
}
autoclass_content = "both"  # Include __init__ in the documentation
autosummary_generate = True  # Generate autosummary


# HTML output
html_theme = "sphinx_rtd_theme"  # One of many possibilities
html_static_path = ["_static"]
html_logo = str(root_path.joinpath("docs", "source", "_static", "logo_b.svg"))



"""Contains the local submitter class."""


import logging

# NOTE: Subprocess below is safe against shell injections
# https://github.com/PyCQA/bandit/issues/280
import subprocess  # nosec
from pathlib import Path

# The import below is also alarming bandit
from subprocess import CompletedProcess  # nosec
from typing import Optional

from bout_runners.submitter.abstract_submitter import AbstractSubmitter
from bout_runners.submitter.processor_split import ProcessorSplit
from bout_runners.utils.file_operations import get_caller_dir


class LocalSubmitter(AbstractSubmitter):
    r"""
    Submits a command.

    Attributes
    ----------
    __pid : None or int
        Getter variable for pid
    path : Path or str
        Directory to run the command from
    processor_split : ProcessorSplit
        Object containing the processor split
    pid : None or int
        The processor id

    Methods
    -------
    submit_command(command)
        Run a subprocess
    _raise_submit_error(self, result):
        Raise and error from the subprocess in a clean way.

    Examples
    --------
    >>> LocalSubmitter().submit_command('ls')
    CompletedProcess(args=['ls'], returncode=0, stdout=b'__init__.py\n
    __pycache__\n
    test_local_submitter.py\n
    test_processor_split.py\n
    test_submitter_factory.py\n', stderr=b'')
    """

    def __init__(
        self, path: Optional[Path] = None, processor_split: None = None
    ) -> None:
        """
        Set the path from where the calls are made from.

        Parameters
        ----------
        path : Path or str or None
            Directory to run the command from
            If None, the calling directory will be used
        processor_split : ProcessorSplit or None
            Object containing the processor split
            If None, default values will be used
        """
        # NOTE: We are not setting the default as a keyword argument
        #       as this would mess up the paths
        self.__path = Path(path).absolute() if path is not None else get_caller_dir()
        self.processor_split = (
            processor_split if processor_split is not None else ProcessorSplit()
        )
        self.__pid: Optional[int] = None

    @property
    def pid(self) -> Optional[int]:
        """
        Return the process id.

        Returns
        -------
        self.__pid : int or None
            The process id if a process has been called, else None
        """
        return self.__pid

    def submit_command(self, command: str) -> CompletedProcess:
        """
        Run a subprocess.

        Parameters
        ----------
        command : str
            The command to run

        Returns
        -------
        result : subprocess.CompletedProcess
            The result of the subprocess call
        """
        logging.info("Executing %s in %s", command, self.__path)
        # This is a simplified subprocess.run(), with the exception
        # that we capture the process id
        process = subprocess.Popen(
            command.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.__path,
            # https://docs.python.org/3/library/subprocess.html#security-considerations
            # https://github.com/PyCQA/bandit/issues/280
            shell=False,  # nosec
        )
        std_out, std_err = process.communicate()
        return_code = process.poll()
        result = subprocess.CompletedProcess(
            process.args, return_code, std_out, std_err
        )
        self.__pid = process.pid

        if result.returncode != 0:
            self._raise_submit_error(result)

        return result

    def _raise_submit_error(self, result: subprocess.CompletedProcess) -> None:
        """
        Raise and error from the subprocess in a clean way.

        Parameters
        ----------
        result : subprocess.CompletedProcess
            The result from the subprocess
        """
        logging.error("Subprocess failed with stdout:")
        logging.error(result.stdout)
        logging.error("and stderr:")
        logging.error(result.stderr)

        result.check_returncode()



"""Module containing file operation functions."""


from datetime import datetime
from pathlib import Path


def get_caller_dir() -> Path:
    """
    Return the directory of the topmost caller file.

    Returns
    -------
    caller_dir : Path
        The path of the topmost caller
    """
    caller_dir = Path().cwd().absolute()

    return caller_dir


def get_modified_time(file_path: Path) -> str:
    """
    Return the modification time of a file path.

    Parameters
    ----------
    file_path : Path
        The file path to get the modification time from

    Returns
    -------
    modified_time : str
        The modification time on ISO8601 format
    """
    # From
    # https://stackoverflow.com/a/52858040/2786884
    modified_time = datetime.fromtimestamp(file_path.stat().st_mtime).isoformat()

    return modified_time



"""Contains unittests for the local submitter."""


# NOTE: subprocess can be vulnerable if shell=True
#       However, CompletedProcess has no known security vulnerabilities
from subprocess import CompletedProcess  # nosec

import pytest
from bout_runners.submitter.local_submitter import LocalSubmitter


def test_local_submitter() -> None:
    """Test that LocalSubmitter can run a command and raise an error."""
    submitter = LocalSubmitter()
    result = submitter.submit_command("ls")
    assert isinstance(result, CompletedProcess)
    assert isinstance(submitter.pid, int)

    with pytest.raises(FileNotFoundError):
        submitter.submit_command("not a real command")



"""Module containing the DatabaseCreator class."""


import logging
import re
from typing import Dict, Optional, Tuple

from bout_runners.database.database_connector import DatabaseConnector
from bout_runners.database.database_utils import get_system_info_as_sql_type


class DatabaseCreator:
    r"""
    Class for creating the schema of the database.

    Attributes
    ----------
    db_connector : DatabaseConnector
        The database object to write to

    Methods
    -------
    create_all_schema_tables(parameters_as_sql_types)
        Create the all the tables for a schema
    get_create_table_statement(table_name, columns=None,
    primary_key='id', foreign_keys=None)
        Return a SQL string which can be used to create the table
    _create_single_table(table_str)
        Create a table in the database
    _create_system_info_table()
        Create a table for the system info
    _create_split_table()
        Create a table which stores the grid split
    _create_file_modification_table()
        Create a table for file modifications
    _create_parameter_tables(parameters_as_sql_types)
        Create a table for each BOUT.settings section and a join table
    _create_run_table()
        Create a table for the metadata_recorder of a run

    Examples
    --------
    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.parameters.default_parameters import DefaultParameters
    >>> from bout_runners.parameters.final_parameters import FinalParameters
    >>> from bout_runners.database.database_connector import DatabaseConnector

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Obtain the parameters

    >>> default_parameters = DefaultParameters(bout_paths)
    >>> final_parameters = FinalParameters(default_parameters)
    >>> final_parameters_dict = final_parameters.get_final_parameters()
    >>> final_parameters_as_sql_types = \
    ...     final_parameters.cast_to_sql_type(final_parameters_dict)

    Create the database

    >>> db_connection = DatabaseConnector('name')
    >>> db_creator = DatabaseCreator(db_connection)
    >>> db_creator.create_all_schema_tables(final_parameters_as_sql_types)
    """

    def __init__(self, db_connector: DatabaseConnector) -> None:
        """
        Set the database to use.

        Parameters
        ----------
        db_connector : DatabaseConnector
            The database object to write to
        """
        self.db_connector = db_connector

    def create_all_schema_tables(
        self, parameters_as_sql_types: Dict[str, Dict[str, str]]
    ) -> None:
        """
        Create the all the tables for a schema.

        The database schema will be on normalized form, see [1]_ for a quick
        overview, and [2]_ for a slightly deeper explanation

        Parameters
        ----------
        parameters_as_sql_types : dict
            The dictionary on the form

            >>> {'section': {'parameter': 'value_type'}}

        References
        ----------
        .. [1] https://www.databasestar.com/database-normalization/
        .. [2] http://www.bkent.net/Doc/simple5.htm
        """
        logging.info("Creating tables in %s", self.db_connector.db_path)

        # Check if tables are created
        self._create_system_info_table()
        self._create_split_table()
        self._create_file_modification_table()
        self._create_parameter_tables(parameters_as_sql_types)
        self._create_run_table()

        logging.info("Tables created in %s", self.db_connector.db_path)

    @staticmethod
    def get_create_table_statement(
        table_name: str,
        columns: Optional[Dict[str, str]] = None,
        primary_key: str = "id",
        foreign_keys: Optional[Dict[str, Tuple[str, str]]] = None,
    ) -> str:
        """
        Return a SQL string which can be used to create the table.

        Parameters
        ----------
        table_name : str
            Name of the table
        columns : dict or None
            Dictionary where the key is the column name and the value is
            the type
        primary_key : str
            Name of the primary key (the type is set to INTEGER)
        foreign_keys : dict or None
            Dictionary where the key is the column in this table to be used as a
            foreign key and the value is the tuple consisting of (name_of_the_table,
            key_in_table) to refer to

        Returns
        -------
        create_statement : str
            The SQL statement which creates table
        """
        create_statement = f"CREATE TABLE {table_name} \n("

        create_statement += f"   {primary_key} INTEGER PRIMARY KEY,\n"

        # These are not known during submission time
        nullable_fields = ("start_time", "stop_time")
        if columns is not None:
            for name, sql_type in columns.items():
                create_statement += f"    {name} {sql_type}"
                if name in nullable_fields:
                    nullable_str = ",\n"
                else:
                    nullable_str = " NOT NULL,\n"
                create_statement += nullable_str

        if foreign_keys is not None:
            # Create the key as column
            # NOTE: All keys are integers
            for name in foreign_keys.keys():
                create_statement += f"    {name} INTEGER NOT NULL,\n"

            # Add constraint
            for name, (f_table_name, key_in_table) in foreign_keys.items():
                # If the parent is updated or deleted, we would like the
                # same effect to apply to its child (thereby the CASCADE
                # parameter)
                create_statement += (
                    f"    FOREIGN KEY({name}) \n"
                    f"        REFERENCES {f_table_name}"
                    f"({key_in_table})\n"
                    f"            ON UPDATE CASCADE\n"
                    f"            ON DELETE CASCADE,"
                    f"\n"
                )

        # Replace last comma with )
        create_statement = f"{create_statement[:-2]})"

        return create_statement

    def _create_single_table(self, table_str: str) -> None:
        """
        Create a table in the database.

        Parameters
        ----------
        table_str : str
            The query to execute

        Raises
        ------
        ValueError
            If the table_str is not understood
        """
        # Obtain the table name
        pattern = r"CREATE TABLE (\w*)"

        match = re.match(pattern, table_str)
        if match is None:
            raise ValueError(f'table_str "{table_str}" not understood')

        table_name = match.group(1)

        self.db_connector.execute_statement(table_str)

        logging.info("Created table %s", table_name)

    def _create_system_info_table(self) -> None:
        """Create a table for the system info."""
        sys_info_dict = get_system_info_as_sql_type()
        sys_info_statement = self.get_create_table_statement(
            table_name="system_info", columns=sys_info_dict
        )
        self._create_single_table(sys_info_statement)

    def _create_split_table(self) -> None:
        """Create a table which stores the grid split."""
        split_statement = self.get_create_table_statement(
            table_name="split",
            columns={
                "number_of_processors": "INTEGER",
                "number_of_nodes": "INTEGER",
                "processors_per_node": "INTEGER",
            },
        )
        self._create_single_table(split_statement)

    def _create_file_modification_table(self) -> None:
        """Create a table for file modifications."""
        file_modification_statement = self.get_create_table_statement(
            table_name="file_modification",
            columns={
                "project_makefile_modified": "TIMESTAMP",
                "project_executable_modified": "TIMESTAMP",
                "project_git_sha": "TEXT",
                "bout_lib_modified": "TIMESTAMP",
                "bout_git_sha": "TEXT",
            },
        )
        self._create_single_table(file_modification_statement)

    def _create_parameter_tables(
        self, parameters_as_sql_types: Dict[str, Dict[str, str]]
    ) -> None:
        """
        Create a table for each BOUT.settings section and a join table.

        Parameters
        ----------
        parameters_as_sql_types : dict
            The dictionary on the form

            >>> {'section': {'parameter': 'value_type'}}

        Notes
        -----
        All `:` will be replaced by `_` in the section names
        """
        parameters_foreign_keys = dict()
        for section in parameters_as_sql_types.keys():
            # Replace bad characters for SQL
            section_name = section.replace(":", "_")
            # Generate foreign keys for the parameters table
            parameters_foreign_keys[f"{section_name}_id"] = (section_name, "id")

            columns = dict()
            for parameter, value_type in parameters_as_sql_types[section].items():
                # Generate the columns
                columns[parameter] = value_type

            # Creat the section table
            section_statement = self.get_create_table_statement(
                table_name=section_name, columns=columns
            )
            self._create_single_table(section_statement)

        # Create the join table
        parameters_statement = self.get_create_table_statement(
            table_name="parameters", foreign_keys=parameters_foreign_keys
        )
        self._create_single_table(parameters_statement)

    def _create_run_table(self) -> None:
        """Create a table for the metadata of a run."""
        run_statement = self.get_create_table_statement(
            table_name="run",
            columns={
                "name": "TEXT",
                "submitted_time": "TIMESTAMP",
                "start_time": "TIMESTAMP",
                "stop_time": "TIMESTAMP",
                "latest_status": "TEXT",
            },
            foreign_keys={
                "file_modification_id": ("file_modification", "id"),
                "split_id": ("split", "id"),
                "parameters_id": ("parameters", "id"),
                "system_info_id": ("system_info", "id"),
            },
        )
        self._create_single_table(run_statement)



"""Contains the class dealing with the final parameters."""


import ast
from typing import Dict, Optional, Union

from bout_runners.parameters.default_parameters import DefaultParameters
from bout_runners.parameters.run_parameters import RunParameters


class FinalParameters:
    r"""
    Class which deals with the final parameters.

    The final parameters are those who are going to be used in the execution of the
    project

    Attributes
    ----------
    self.__default_parameters : DefaultParameters
        Object dealing with the default parameters
    self.__run_parameters = run_parameters
        Object dealing with the run parameters

    Methods
    -------
    get_final_parameters()
        Obtain the final parameters that will be used in a run
    cast_parameters_to_sql_type(parameter_dict)
        Cast the values of a parameter dict to valid SQL types

    Examples
    --------
    The easiest way to use FinalParameters is to run a script from the root directory
    of the project (i.e. where the `Makefile` and `data` directory are normally
    situated. The script can simply call

    >>> FinalParameters().get_final_parameters()
    {'global': {'append': False, 'async_send': False, ...}}

    A more elaborate example where all the dependency objects are
    built manually:

    Import dependencies

    >>> from pathlib import Path
    >>> from bout_runners.executor.bout_paths import BoutPaths
    >>> from bout_runners.parameters.default_parameters import DefaultParameters
    >>> from bout_runners.parameters.run_parameters import RunParameters

    Create the `bout_paths` object

    >>> project_path = Path().joinpath('path', 'to', 'project')
    >>> bout_inp_src_dir = Path().joinpath('path', 'to', 'source', 'BOUT.inp')
    >>> bout_inp_dst_dir = Path().joinpath('path', 'to', 'destination', 'BOUT.inp')
    >>> bout_paths = BoutPaths(project_path=project_path,
    ...                        bout_inp_src_dir=bout_inp_src_dir,
    ...                        bout_inp_dst_dir=bout_inp_dst_dir)

    Get the final parameters

    >>> default_parameters = DefaultParameters(bout_paths=bout_paths)
    >>> final_parameters = FinalParameters(default_parameters,
    ...     RunParameters({'nout': 10}))
    >>> final_parameters_dict = final_parameters.get_final_parameters()
    >>> final_parameters_dict
    {'global': {'append': False, 'async_send': False, ..., 'nout': 10, ...}}

    >>> final_parameters.\
    ...     cast_to_sql_type(final_parameters_dict)
    {'global': {'append': TEXT, 'async_send': TETX, ..., 'nout': INTEGER, ...}}
    """

    def __init__(
        self,
        default_parameters: Optional[DefaultParameters] = None,
        run_parameters: Optional[RunParameters] = None,
    ) -> None:
        """
        Set the member data.

        Parameters
        ----------
        default_parameters : DefaultParameters or None
            Object dealing with default parameters (i.e. standard BOUT++ parameters,
            or those given in BOUT.inp)
        run_parameters : RunParameters or None
            Object dealing with run parameters (i.e. parameters set in bout_runner
            which has precedence over BOUT.inp)
            If None, default parameters will be used
        """
        self.__default_parameters = (
            default_parameters
            if default_parameters is not None
            else DefaultParameters()
        )
        self.__run_parameters = (
            run_parameters if run_parameters is not None else RunParameters()
        )

    def get_final_parameters(
        self,
    ) -> Dict[str, Dict[str, Union[str, int, float, bool]]]:
        """
        Obtain the final parameters that will be used in a run.

        Returns
        -------
        final_parameters_dict : dict
            Parameters on the form

            >>> {'global':{'append': 'False', 'nout': 5},
            ...  'mesh':  {'nx': 4},
            ...  'section_in_BOUT_inp': {'some_variable': 'some_value'}}

        Raises
        ------
        RuntimeError
            If run_parameters_dict is None
        """
        final_parameters_dict = self.__default_parameters.get_default_parameters()
        run_parameters_dict = self.__run_parameters.run_parameters_dict
        # Assert to prevent "Incompatible types in assignment" with Optional
        if run_parameters_dict is None:
            raise RuntimeError("run_parameters_dict is None")
        final_parameters_dict.update(run_parameters_dict)

        # Cast True to 1 and False to 0 as SQLite has no support for
        # bool
        sections = final_parameters_dict.keys()
        for section in sections:
            parameters = final_parameters_dict[section].keys()
            for parameter in parameters:
                val = final_parameters_dict[section][parameter]
                if isinstance(val, bool):
                    if val:
                        final_parameters_dict[section][parameter] = 1
                    else:
                        final_parameters_dict[section][parameter] = 0

        return final_parameters_dict

    @staticmethod
    def cast_to_sql_type(
        parameter_dict: Dict[str, Dict[str, Union[str, int, float]]]
    ) -> Dict[str, Dict[str, str]]:
        """
        Cast the values of a parameter dict to valid SQL types.

        Parameters
        ----------
        parameter_dict : dict
            Dictionary containing the parameters given in BOUT.settings
            On the form

            >>> {'section': {'parameter': 'value'}}

        Returns
        -------
        parameter_dict_as_sql_types : dict
            Dictionary containing the parameters given in BOUT.settings
            On the form

            >>> {'section': {'parameter': 'value_type'}}
        """
        type_map = {
            "bool": "INTEGER",  # No bool type in SQLite
            "float": "REAL",
            "int": "INTEGER",
            "str": "TEXT",
        }

        parameter_dict_copy = parameter_dict.copy()
        parameter_dict_as_sql_types: Dict[str, Dict[str, str]] = dict()

        for section in parameter_dict_copy.keys():
            parameter_dict_as_sql_types[section] = dict()
            for key, val in parameter_dict_copy[section].items():
                # If type is not found, type is str
                try:
                    val_type = type(ast.literal_eval(str(val)))
                except (SyntaxError, ValueError):
                    val_type = str

                parameter_dict_as_sql_types[section][key] = type_map[val_type.__name__]

        return parameter_dict_as_sql_types



"""Contains unittests for the executor."""


from pathlib import Path
from typing import Callable

from bout_runners.parameters.default_parameters import DefaultParameters


def test_executor(make_project: Path, yield_bout_path_conduction: Callable) -> None:
    """
    Test that we are able to execute the conduction example.

    Parameters
    ----------
    make_project : Path
        The path to the conduction example
    yield_bout_path_conduction : function
        Function which makes the BoutPaths object for the conduction example
    """
    # Use the make fixture in order to automate clean up after done
    _ = make_project

    # Make the executor
    bout_paths = yield_bout_path_conduction("test_executor")
    executor = DefaultParameters.get_test_executor(bout_paths)

    executor.execute()

    log_path = bout_paths.bout_inp_dst_dir.joinpath("BOUT.log.0")

    assert log_path.is_file()



"""Contains unittests for the SubmitterFactory."""


import pytest
from bout_runners.submitter.local_submitter import LocalSubmitter
from bout_runners.submitter.submitter_factory import get_submitter


def test_submitter_factory() -> None:
    """Test that the SubmitterFactory returns Submitter objects."""
    submitter = get_submitter("local")
    assert isinstance(submitter, LocalSubmitter)

    with pytest.raises(NotImplementedError):
        get_submitter("not a class")



"""Contains unittests for run parameters."""


from bout_runners.parameters.run_parameters import RunParameters


def test_run_parameters() -> None:
    """Test that the RunParameters is setting the parameters."""
    run_parameters = RunParameters({"global": {"append": False}, "mesh": {"nx": 4}})
    expected_str = "append=False mesh.nx=4 "
    assert run_parameters.run_parameters_str == expected_str



"""Module containing the classes to read the makefile."""


import re
from pathlib import Path


class MakefileReaderError(Exception):
    """Error class indicating that this is a ReadMakefile error."""

    def __init__(self, variable: str, path: Path) -> None:
        """
        Construct a string and call the super constructor.

        Parameters
        ----------
        variable : str
            Variable to searched for
        path : Path or str
            Path searched at
        """
        message = f"Could not find {variable} in {path}"

        super().__init__(message)


class BoutMakefileReader:
    """
    Class which reads a BOUT++ Makefile.

    Attributes
    ----------
    path : Path
        The path to the Makefile
    content : str
        The content of the Makefile as a string
    """

    def __init__(self, path: Path) -> None:
        """
        Read the content of a Makefile and store it into self.content.

        Parameters
        ----------
        path : Path or str
            Path to the Makefile
        """
        self.path = path
        self.content = self.read()

    def read(self) -> str:
        """
        Read the makefile.

        Returns
        -------
        str
            Content of the Makefile
        """
        with Path(self.path).open("r") as make_file:
            return make_file.read()

    @property
    def value(self) -> str:
        """Get the value of the variable."""
        return ""


class BoutMakefileVariableReader(BoutMakefileReader):
    r"""
    Class which reads a variable from a BOUT++ Makefile.

    Attributes
    ----------
    variable_name : str
        Name of the variable belonging to the instance
    value : str
        Value belonging to the variable of the instance

    Methods
    -------
    get_variable_value()
        Get the value of the variable

    Examples
    --------
    Makefile

    >>> BOUT_SUPER	= /super/path/to/BOUT-dev
    ... BOUT_TOP	= $(BOUT_SUPER)/BOUT-dev
    ...
    ... SOURCEC		= bout_model.cxx
    ...
    ... include $(BOUT_TOP)/make.config

    Script

    >>> BoutMakefileVariableReader('SOURCEC', 'Makefile').value
    'bout_model.cxx'
    """

    def __init__(self, path: Path, variable_name: str) -> None:
        """
        Set the variable name of the instance.

        Parameters
        ----------
        path : Path or str
            The path to the Makefile
        variable_name : str
            The variable under consideration
        """
        super(BoutMakefileVariableReader, self).__init__(path)

        self.variable_name = variable_name

    @property
    def value(self) -> str:
        """
        Get the value of the variable.

        Returns
        -------
        value : str
            The last match of the variable

        Raises
        ------
        MakefileReaderError
            If self.variable is not found

        Examples
        --------
        Makefile

        >>> # foo=bar
        ... foo = baz
        ... foo   = foobar.qux # foo = quux.quuz

        Script

        >>> BoutMakefileVariableReader('foo', 'Makefile').value
        'foobar.qux'
        """
        # Build the match function for the regex
        no_comment_line = r"^\s*(?!#)"
        must_contain_eq_sign = r"\s*=\s*"
        # As we are using the MULTILINE modifier we should exclude
        # newlines
        capture_all_except_comment_and_newline = r"([.]*[^#\n]*)"
        avoid_trailing_whitespace = r"(?<!\s)"

        pattern = (
            f"{no_comment_line}{self.variable_name}"
            f"{must_contain_eq_sign}"
            f"{capture_all_except_comment_and_newline}"
            f"{avoid_trailing_whitespace}"
        )

        # NOTE: match() checks for match only at the beginning of a
        #       string
        #       search() checks for match anywhere in the string,
        #       but returns only the first occurrence
        #       findall() matches all matching group as a list,
        #       and as we only have one matching group in the
        #       pattern the elements in the list will be string not a
        #       tuple
        matches = re.findall(pattern, self.content, re.M)

        if len(matches) == 0:
            raise MakefileReaderError(self.variable_name, self.path)

        # Only the last line of the variable will be considered by
        # the Makefile
        value: str = matches[-1]

        return value



"""Contains unittests for the database utils."""

from bout_runners.database.database_utils import get_system_info_as_sql_type


def test_get_system_info_as_sql_type() -> None:
    """Test that the system info can be returned as a dict."""
    sys_info_dict = get_system_info_as_sql_type()
    assert isinstance(sys_info_dict, dict)



"""Contains unittests for make."""


from pathlib import Path
from typing import Tuple

from bout_runners.make.make import Make


def test_make_project(make_make_object: Tuple[Make, Path]) -> None:
    """
    Test that the MakeProject class is able to make conduction.

    Parameters
    ----------
    make_make_object : tuple
        Tuple consisting of the `make_obj` (MakeProject object) and `exec_file` (Path)
        See the make_make_object fixture for details

    See Also
    --------
    make_make_object : Fixture which makes the make object
    """
    make_obj, exec_file = make_make_object

    # NOTE: The setup runs make clean, so the project directory
    #       should not contain any executable
    assert not exec_file.is_file()

    make_obj.run_make()

    assert exec_file.is_file()

    # Check that the file is not made again
    # For detail, see
    # https://stackoverflow.com/a/52858040/2786884
    first_creation_time = exec_file.stat().st_ctime

    make_obj.run_make()

    assert first_creation_time == exec_file.stat().st_ctime

    # Check that the force flag makes the project again
    make_obj.run_make(force=True)

    assert first_creation_time != exec_file.stat().st_ctime
